import asyncio
import cgi
import concurrent.futures
import email.utils
import inspect
import multiprocessing
import os
import os.path
import pprint
import re
import stat
import subprocess
import time
from concurrent.futures import ProcessPoolExecutor
from copy import deepcopy
from datetime import datetime, timezone
from ftplib import FTP
from functools import partial
from typing import Any, Callable, Dict, Generator, Iterable, List, Optional, Tuple, Union
from urllib import parse as urlparse

try:
    import docker
    from docker.errors import ImageNotFound, NotFound, NullResource

    docker_avail = True
except ImportError:
    docker_avail = False

import orjson
import requests

from biothings import config as btconfig
from biothings.hub import DUMPER_CATEGORY, UPLOADER_CATEGORY, renderer as job_renderer
from biothings.hub.dataload.uploader import set_pending_to_upload
from biothings.utils.common import rmdashfr, timesofar, untarall
from biothings.utils.hub_db import get_src_dump
from biothings.utils.loggers import get_logger
from biothings.utils.manager import BaseSourceManager, ResourceError
from biothings.utils.parsers import docker_source_info_parser

logging = btconfig.logger


class DumperException(Exception):
    pass


class DockerContainerException(Exception):
    pass


class BaseDumper(object):
    # override in subclass accordingly
    SRC_NAME = None
    SRC_ROOT_FOLDER = None  # source folder (without version/dates)

    # Should an upload be triggered after dump ?
    AUTO_UPLOAD = True

    # attribute used to generate data folder path suffix
    # by default, it uses self.release value, which is typically set in self.set_release method
    SUFFIX_ATTR = "release"

    # Max parallel downloads (None = no limit).
    MAX_PARALLEL_DUMP = None
    # waiting time between download (0.0 = no waiting)
    SLEEP_BETWEEN_DOWNLOAD = 0.0

    # keep all release (True) or keep only the latest ?
    ARCHIVE = True

    SCHEDULE = None  # crontab format schedule, if None, won't be scheduled

    def __init__(self, src_name=None, src_root_folder=None, log_folder=None, archive=None):
        # unpickable attrs, grouped
        self.init_state()
        self.src_name = src_name or self.SRC_NAME
        self.src_root_folder = src_root_folder or self.SRC_ROOT_FOLDER
        self.log_folder = log_folder or btconfig.LOG_FOLDER
        self.archive = archive or self.ARCHIVE
        self.to_dump = []
        self.to_delete: List[Union[str, bytes, os.PathLike]] = []
        """Populate with list of relative path of files to delete"""
        self.release = None
        self.t0 = time.time()
        self.logfile = None
        self.prev_data_folder = None
        self.timestamp = time.strftime("%Y%m%d")
        self.prepared = False
        self.steps = ["dump", "post"]

    def init_state(self):
        self._state = {
            "client": None,
            "src_dump": None,
            "logger": None,
            "src_doc": None,
        }

    # specific setters for attrs that can't be pickled
    # note: we can't use a generic __setattr__ as it collides
    # (infinite recursion) with __getattr__, and we can't use
    # __getattr__ as well as @x.setter required @property(x) to
    # be defined. We'll be explicit there...
    @property
    def client(self):
        if not self._state["client"]:
            try:
                self.prepare_client()
            except Exception as e:
                # if accessed but not ready, then just ignore and return invalid value for a client
                logging.exception("Can't prepare client: %s" % e)
                return None
        return self._state["client"]

    @property
    def src_dump(self):
        if not self._state["src_dump"]:
            self.prepare()
        return self._state["src_dump"]

    @property
    def logger(self):
        if not self._state["logger"]:
            self.prepare()
        return self._state["logger"]

    @property
    def src_doc(self):
        # this one is pickable but it's a lazy load
        # (based on non-pickable src_dump)
        if not self._state["src_doc"]:
            self.prepare()
        return self._state["src_doc"]

    @client.setter
    def client(self, value):
        self._state["client"] = value

    @src_dump.setter
    def src_dump(self, value):
        self._state["src_dump"] = value

    @logger.setter
    def logger(self, value):
        self._state["logger"] = value

    @src_doc.setter
    def src_doc(self, value):
        self._state["src_doc"] = value

    def create_todump_list(self, force=False, **kwargs):
        """Fill self.to_dump list with dict("remote":remote_path,"local":local_path)
        elements. This is the todo list for the dumper. It's a good place to
        check whether needs to be downloaded. If 'force' is True though, all files
        will be considered for download"""
        raise NotImplementedError("Define in subclass")

    def prepare_client(self):
        """do initialization to make the client ready to dump files"""
        raise NotImplementedError("Define in subclass")

    def need_prepare(self):
        """check whether some prepare step should executed
        before running dump"""

    def release_client(self):
        """Do whatever necessary (like closing network connection)
        to "release" the client"""
        raise NotImplementedError("Define in subclass")

    def remote_is_better(self, remotefile, localfile):
        """Compared to local file, check if remote file is worth downloading.
        (like either bigger or newer for instance)"""
        raise NotImplementedError("Define in subclass")

    def download(self, remotefile, localfile):
        """
        Download "remotefile' to local location defined by 'localfile'
        Return relevant information about remotefile (depends on the actual client)
        """
        raise NotImplementedError("Define in subclass")

    def post_download(self, remotefile, localfile):
        """Placeholder to add a custom process once a file is downloaded.
        This is a good place to check file's integrity. Optional"""
        pass

    def post_dump_delete_files(self):
        """
        Delete files after dump

        Invoke this method in post_dump to synchronously delete
        the list of paths stored in `self.to_delete`, in order.

        Non-recursive. If directories need to be removed, build the list such that
        files residing in the directory are removed first and then the directory.
        (Hint: see `os.walk(dir, topdown=False)`)
        """
        base_dir: str = os.path.realpath(self.new_data_folder)
        self.logger.debug("Only delete files under %s", base_dir)
        # assume this path is good
        for rel_file_name in self.to_delete:
            delete_path = os.path.realpath(os.path.join(base_dir, rel_file_name))  # figure out the full path
            self.logger.debug("%s is %s", rel_file_name, delete_path)
            common_path = os.path.commonpath((base_dir, delete_path))
            self.logger.debug("Calculated common prefix path: %s", common_path)
            if common_path != base_dir or delete_path == base_dir:
                raise RuntimeError("Attempting to delete something outside the download " "directory")
            try:
                s = os.stat(delete_path)
                self.logger.debug("stat(%s): %s", delete_path, s)
            except FileNotFoundError:
                self.logger.warning("Cannot delete %s (%s), does not exist", rel_file_name, delete_path)
                continue
            # there is a race condition but the effects are limited
            if stat.S_ISREG(s.st_mode):
                self.logger.info("Deleting regular file %s (%s)", rel_file_name, delete_path)
                try:
                    os.unlink(delete_path)
                except Exception as e:
                    self.logger.exception("Failed to delete regular file")
                    raise e
            elif stat.S_ISDIR(s.st_mode):
                self.logger.info("Deleting directory %s (%s)", rel_file_name, delete_path)
                try:
                    os.rmdir(delete_path)
                except Exception as e:
                    self.logger.exception("Failed to delete directory")
                    raise e
            else:
                raise RuntimeError(
                    f"{rel_file_name} ({delete_path}) is not " "a regular file or directory, cannot delete"
                )
        self.to_delete = []  # reset the list in case

    def post_dump(self, *args, **kwargs):
        """
        Placeholder to add a custom process once the whole resource
        has been dumped. Optional.
        """
        pass

    def setup_log(self):
        log_folder = os.path.join(btconfig.LOG_FOLDER, "dataload") if btconfig.LOG_FOLDER else None
        self.logger, self.logfile = get_logger("dump_%s" % self.src_name, log_folder=log_folder)

    def prepare(self, state={}):  # noqa: B006
        if self.prepared:
            return
        if state:
            # let's be explicit, _state takes what it wants
            for k in self._state:
                self._state[k] = state[k]
            return
        self.prepare_src_dump()
        self.setup_log()

    def unprepare(self):
        """
        reset anything that's not pickable (so self can be pickled)
        return what's been reset as a dict, so self can be restored
        once pickled
        """
        state = {
            "client": self._state["client"],
            "src_dump": self._state["src_dump"],
            "logger": self._state["logger"],
            "src_doc": self._state["src_doc"],
        }
        for k in state:
            self._state[k] = None
        self.prepared = False
        return state

    def prepare_src_dump(self):
        # Mongo side
        self.src_dump = get_src_dump()
        self.src_doc = self.src_dump.find_one({"_id": self.src_name}) or {}

    def register_status(self, status, transient=False, dry_run=False, **extra):
        src_doc = deepcopy(self.src_doc)
        try:
            # if status is "failed" and depending on where it failed,
            # we may not be able to get the new_data_folder (if dumper didn't reach
            # the release information for instance). Default to current if failing
            data_folder = self.new_data_folder
        except DumperException:
            data_folder = self.current_data_folder
        release = getattr(self, self.__class__.SUFFIX_ATTR)
        if release is None:
            # it has not been set by the dumper before while exploring
            # remote site. maybe we're just running post step ?
            # back-compatibility; use "release" at root level if not found under "download"
            release = src_doc.get("download", {}).get("release") or src_doc.get("release")
            self.logger.error("No release set, assuming: data_folder: %s, release: %s" % (data_folder, release))
        # make sure to remove old "release" field to get back on track
        for field in ["release", "data_folder"]:
            if src_doc.get(field):
                self.logger.warning("Found '%s'='%s' at root level, convert to new format" % (field, src_doc[field]))
                src_doc.pop(field)

        current_download_info = {
            "_id": self.src_name,
            "download": {
                "release": release,
                "data_folder": data_folder,
                "logfile": self.logfile,
                "started_at": datetime.now().astimezone(),
                "status": status,
            },
        }
        # Update last success download time.
        if status == "success":
            # If current status is success, we will get the current's started_at
            last_success = current_download_info["download"]["started_at"]
        else:
            # If failed, we will get the last_success from the last download instead.
            last_download_info = src_doc.setdefault("download", {})
            last_success = last_download_info.get("last_success", None)
            if not last_success and last_download_info.get("status") == "success":
                # If last_success from the last download doesn't exist or is None, and last
                # download's status is success, the last download's started_at will be used.
                last_success = last_download_info.get("started_at")
        if last_success:
            current_download_info["download"]["last_success"] = last_success

        src_doc.update(current_download_info)

        # only register time when it's a final state
        if transient:
            src_doc["download"]["pid"] = os.getpid()
        else:
            src_doc["download"]["time"] = timesofar(self.t0)
        if "download" in extra:
            src_doc["download"].update(extra["download"])
        else:
            src_doc.update(extra)

        # when dry run, we should not change the src_doc, and src_dump
        if not dry_run:
            self.src_doc = deepcopy(src_doc)
            self.src_dump.save(src_doc)

    async def dump(self, steps=None, force=False, job_manager=None, check_only=False, **kwargs):
        """
        Dump (ie. download) resource as needed
        this should be called after instance creation
        'force' argument will force dump, passing this to
        create_todump_list() method.
        """
        # signature says it's optional but for now it's not...
        assert job_manager
        # check what to do
        self.steps = steps or self.steps
        if type(self.steps) == str:
            self.steps = [self.steps]
        strargs = "[steps=%s]" % ",".join(self.steps)
        try:
            if "dump" in self.steps:
                pinfo = self.get_pinfo()
                pinfo["step"] = "check"
                # if last download failed (or was interrupted), we want to force the dump again
                try:
                    if self.src_doc["download"]["status"] in ["failed", "downloading"]:
                        self.logger.info("Forcing dump because previous failed (so let's try again)")
                        force = True
                except (AttributeError, KeyError):
                    # no src_doc or no download info
                    pass
                # TODO: blocking call for now, FTP client can't be properly set in thread after
                if inspect.iscoroutinefunction(self.create_todump_list):
                    await self.create_todump_list(force=force, job_manager=job_manager, **kwargs)
                else:
                    self.create_todump_list(force=force, **kwargs)
                # make sure we release (disconnect) client so we don't keep an open
                # connection for nothing
                self.release_client()
                if self.to_dump:
                    if check_only:
                        self.logger.info(
                            "New release available, '%s', %s file(s) to download" % (self.release, len(self.to_dump)),
                            extra={"notify": True},
                        )
                        return self.release
                    # mark the download starts
                    self.register_status("downloading", transient=True)
                    # unsync to make it pickable
                    state = self.unprepare()
                    await self.do_dump(job_manager=job_manager)
                    # then restore state
                    self.prepare(state)
                else:
                    # if nothing to dump, don't do post process
                    self.logger.debug("Nothing to dump", extra={"notify": True})
                    return "Nothing to dump"
            if "post" in self.steps:
                got_error = False
                pinfo = self.get_pinfo()
                pinfo["step"] = "post_dump"
                # for some reason (like maintaining object's state between pickling).
                # we can't use process there. Need to use thread to maintain that state without
                # building an unmaintainable monster
                job = await job_manager.defer_to_thread(pinfo, partial(self.post_dump, job_manager=job_manager))

                def postdumped(f):
                    nonlocal got_error
                    if f.exception():
                        got_error = f.exception()

                job.add_done_callback(postdumped)
                await job
                if got_error:
                    raise got_error
                # set it to success at the very end
                self.register_status("success")
                if self.__class__.AUTO_UPLOAD:
                    set_pending_to_upload(self.src_name)
                self.logger.info("success %s" % strargs, extra={"notify": True})
        except (KeyboardInterrupt, Exception) as e:
            self.logger.error("Error while dumping source: %s" % e)
            import traceback

            self.logger.error(traceback.format_exc())
            self.register_status("failed", download={"err": str(e), "tb": traceback.format_exc()})
            self.logger.error("failed %s: %s" % (strargs, e), extra={"notify": True})
            raise
        finally:
            if self.client:
                self.release_client()

    def mark_success(self, dry_run=True):
        """
        Mark the datasource as successful dumped.
        It's useful in case the datasource is unstable, and need to be manually downloaded.
        """
        self.register_status("success", dry_run=dry_run)
        self.logger.info("Done!")
        result = {
            "_id": self.src_doc["_id"],
            "download": self.src_doc["download"],
        }
        return result

    def get_predicates(self):
        """
        Return a list of predicates (functions returning true/false, as in math logic)
        which instructs/dictates if job manager should start a job (process/thread)
        """

        def no_corresponding_uploader_running(job_manager):
            """
            Don't download data if the associated uploader is running
            """
            return (
                len(
                    [
                        j
                        for j in job_manager.jobs.values()
                        if j["source"].split(".")[0] == self.src_name and j["category"] == UPLOADER_CATEGORY
                    ]
                )
                == 0
            )

        return [no_corresponding_uploader_running]

    def get_pinfo(self):
        """
        Return dict containing information about the current process
        (used to report in the hub)
        """
        pinfo = {
            "category": DUMPER_CATEGORY,
            "source": self.src_name,
            "step": None,
            "description": None,
        }
        preds = self.get_predicates()
        if preds:
            pinfo["__predicates__"] = preds
        return pinfo

    @property
    def new_data_folder(self):
        """Generate a new data folder path using src_root_folder and
        specified suffix attribute. Also sync current (aka previous) data
        folder previously registeted in database.
        This method typically has to be called in create_todump_list()
        when the dumper actually knows some information about the resource,
        like the actual release.
        """
        if self.archive:
            if getattr(self, self.__class__.SUFFIX_ATTR) is None:  # defined but not set
                # if step is "post" only, it means we didn't even check a new version and we
                # want to run "post" step on current version again
                if self.steps == ["post"]:
                    return (
                        self.current_data_folder
                    )  # FIXME: this causes a recursive error as current_data_folder calls new_data_folder too in some cases
                else:
                    raise DumperException(
                        "Can't generate new data folder, attribute used for suffix (%s) isn't set"
                        % self.__class__.SUFFIX_ATTR
                    )
            suffix = getattr(self, self.__class__.SUFFIX_ATTR)
            return os.path.join(self.src_root_folder, suffix)
        else:
            return os.path.join(self.src_root_folder, "latest")

    @property
    def current_data_folder(self):
        try:
            return self.src_doc.get("download", {}).get("data_folder") or self.new_data_folder
        except DumperException:
            # exception raied from new_data_folder generation, we give up
            return None

    @property
    def current_release(self):
        return self.src_doc.get("download", {}).get("release")

    async def do_dump(self, job_manager=None):
        self.logger.info("%d file(s) to download" % len(self.to_dump))
        # should downloads be throttled ?
        max_dump = self.__class__.MAX_PARALLEL_DUMP and asyncio.Semaphore(self.__class__.MAX_PARALLEL_DUMP)
        courtesy_wait = self.__class__.SLEEP_BETWEEN_DOWNLOAD
        got_error = None
        jobs = []
        self.unprepare()
        for todo in self.to_dump:
            remote = todo["remote"]
            local = todo["local"]

            def done(f):
                try:
                    _ = f.result()
                    nonlocal max_dump
                    nonlocal got_error
                    if max_dump:
                        # self.logger.debug("Releasing download semaphore: %s" % max_dump)
                        max_dump.release()
                    self.post_download(remote, local)
                except Exception as e:
                    self.logger.exception("Error downloading '%s': %s", remote, e)
                    got_error = e

            pinfo = self.get_pinfo()
            pinfo["step"] = "dump"
            pinfo["description"] = remote
            if max_dump:
                await max_dump.acquire()
            if courtesy_wait:
                await asyncio.sleep(courtesy_wait)
            job = await job_manager.defer_to_process(pinfo, partial(self.download, remote, local))
            job.add_done_callback(done)
            jobs.append(job)
            # raise error as soon as we get it:
            # 1. it prevents from launching things for nothing
            # 2. if we gather the error at the end of the loop *and* if we
            #    have more errors than the queue size, we get stuck
            if got_error:
                raise got_error
        await asyncio.gather(*jobs)
        if got_error:
            raise got_error
        self.logger.info("%s successfully downloaded" % self.SRC_NAME)
        self.to_dump = []

    def prepare_local_folders(self, localfile):
        localdir = os.path.dirname(localfile)
        if not os.path.exists(localdir):
            try:
                os.makedirs(localdir)
            except FileExistsError:
                # ignore, might exist now (parallelization occuring...)
                pass


class FTPDumper(BaseDumper):
    FTP_HOST = ""
    CWD_DIR = ""
    FTP_USER = ""
    FTP_PASSWD = ""
    FTP_TIMEOUT = 10 * 60.0  # we want dumper to timout if necessary
    BLOCK_SIZE: Optional[int] = None  # default is still kept at 8KB

    # TODO: should we add a __del__ to make sure to close ftp connection ?
    # ftplib has a context __enter__, but we don't use it that way ("with ...")

    def _get_optimal_buffer_size(self) -> int:
        if self.BLOCK_SIZE is not None:
            return self.BLOCK_SIZE
        # else:
        known_optimal_sizes = {
            "ftp.ncbi.nlm.nih.gov": 33554432,
            # see https://ftp.ncbi.nlm.nih.gov/README.ftp for reason
            # add new ones above
            "DEFAULT": 8192,
        }
        normalized_host = self.FTP_HOST.lower()
        if normalized_host in known_optimal_sizes:
            return known_optimal_sizes[normalized_host]
        else:
            return known_optimal_sizes["DEFAULT"]

    def prepare_client(self):
        # FTP side
        self.client = FTP(self.FTP_HOST, timeout=self.FTP_TIMEOUT)
        self.client.login(self.FTP_USER, self.FTP_PASSWD)
        if self.CWD_DIR:
            self.client.cwd(self.CWD_DIR)

    def need_prepare(self):
        return not self.client or (self.client and not self.client.file)

    def release_client(self):
        assert self.client
        self.client.close()
        self.client = None

    def download(self, remotefile, localfile):
        self.prepare_local_folders(localfile)
        self.logger.debug("Downloading '%s' as '%s'" % (remotefile, localfile))
        block_size = self._get_optimal_buffer_size()
        if self.need_prepare():
            self.prepare_client()
        try:
            with open(localfile, "wb") as out_f:
                self.client.retrbinary(cmd="RETR %s" % remotefile, callback=out_f.write, blocksize=block_size)
            # set the mtime to match remote ftp server
            response = self.client.sendcmd("MDTM " + remotefile)
            code, lastmodified = response.split()
            # an example: 'last-modified': '20121128150000'
            lastmodified = time.mktime(datetime.strptime(lastmodified, "%Y%m%d%H%M%S").timetuple())
            os.utime(localfile, (lastmodified, lastmodified))
            return code
        except Exception as e:
            self.logger.error("Error while downloading %s: %s" % (remotefile, e))
            self.release_client()
            raise
        finally:
            self.release_client()

    def remote_is_better(self, remotefile, localfile):
        """'remotefile' is relative path from current working dir (CWD_DIR),
        'localfile' is absolute path"""
        try:
            res = os.stat(localfile)
        except FileNotFoundError:
            # no local file, remote is always better
            return True
        local_lastmodified = int(res.st_mtime)
        self.logger.info("Getting modification time for '%s'" % remotefile)
        response = self.client.sendcmd("MDTM " + remotefile)
        code, remote_lastmodified = response.split()
        remote_lastmodified = int(time.mktime(datetime.strptime(remote_lastmodified, "%Y%m%d%H%M%S").timetuple()))

        if remote_lastmodified > local_lastmodified:
            self.logger.debug(
                "Remote file '%s' is newer (remote: %s, local: %s)"
                % (remotefile, remote_lastmodified, local_lastmodified)
            )
            return True
        local_size = res.st_size
        self.client.sendcmd("TYPE I")
        response = self.client.sendcmd("SIZE " + remotefile)
        code, remote_size = map(int, response.split())
        if remote_size > local_size:
            self.logger.debug(
                "Remote file '%s' is bigger (remote: %s, local: %s)" % (remotefile, remote_size, local_size)
            )
            return True
        self.logger.debug("'%s' is up-to-date, no need to download" % remotefile)
        return False


class LastModifiedBaseDumper(BaseDumper):
    """
    Use SRC_URLS as a list of URLs to download and
    implement create_todump_list() according to that list.
    Shoud be used in parallel with a dumper talking the
    actual underlying protocol
    """

    SRC_URLS = []  # must be overridden in subclass

    def set_release(self):
        """
        Set self.release attribute as the last-modified datetime found in
        the last SRC_URLs element (so releae is the datetime of the last file
        to download)
        """
        raise NotImplementedError("Implement me in sub-class")

    def create_todump_list(self, force=False):
        assert type(self.__class__.SRC_URLS) is list, "SRC_URLS should be a list"
        assert self.__class__.SRC_URLS, "SRC_URLS list is empty"
        self.set_release()  # so we can generate new_data_folder
        for src_url in self.__class__.SRC_URLS:
            filename = os.path.basename(src_url)
            new_localfile = os.path.join(self.new_data_folder, filename)
            try:
                current_localfile = os.path.join(self.current_data_folder, filename)
            except TypeError:
                # current data folder doesn't even exist
                current_localfile = new_localfile

            remote_better = self.remote_is_better(src_url, current_localfile)
            if force or current_localfile is None or remote_better:
                new_localfile = os.path.join(self.new_data_folder, filename)
                self.to_dump.append({"remote": src_url, "local": new_localfile})


class LastModifiedFTPDumper(LastModifiedBaseDumper):
    """
    SRC_URLS containing a list of URLs pointing to files to download,
    use FTP's MDTM command to check whether files should be downloaded
    The release is generated from the last file's MDTM in SRC_URLS, and
    formatted according to RELEASE_FORMAT.
    See also LastModifiedHTTPDumper, working the same way but for HTTP
    protocol.
    Note: this dumper is a wrapper over FTPDumper, one URL will give
    one FTPDumper instance.
    """

    RELEASE_FORMAT = "%Y-%m-%d"

    def prepare_client(self):
        pass

    def release_client(self):
        pass

    def get_client_for_url(self, url):
        split = urlparse.urlsplit(url)
        klass = type(
            "dynftpdumper",
            (FTPDumper,),
            {
                "FTP_HOST": split.hostname,
                "CWD_DIR": "/".join(split.path.split("/")[:-1]),
                "FTP_USER": split.username or "",
                "FTP_PASSWD": split.password or "",
                "SRC_NAME": self.__class__.SRC_NAME,
                "SRC_ROOT_FOLDER": self.__class__.SRC_ROOT_FOLDER,
            },
        )
        ftpdumper = klass()
        ftpdumper.prepare_client()
        return ftpdumper

    def get_remote_file(self, url):
        split = urlparse.urlsplit(url)
        remotef = split.path.split("/")[-1]
        return remotef

    def set_release(self):
        url = self.__class__.SRC_URLS[-1]
        ftpdumper = self.get_client_for_url(url)
        remotefile = self.get_remote_file(url)
        response = ftpdumper.client.sendcmd("MDTM " + remotefile)
        code, lastmodified = response.split()
        lastmodified = time.mktime(datetime.strptime(lastmodified, "%Y%m%d%H%M%S").timetuple())
        dt = datetime.fromtimestamp(lastmodified)
        self.release = dt.strftime(self.__class__.RELEASE_FORMAT)
        ftpdumper.release_client()

    def remote_is_better(self, urlremotefile, localfile):
        ftpdumper = self.get_client_for_url(urlremotefile)
        remotefile = self.get_remote_file(urlremotefile)
        isitbetter = ftpdumper.remote_is_better(remotefile, localfile)
        ftpdumper.release_client()
        return isitbetter

    def download(self, urlremotefile, localfile, headers={}):  # noqa: B006
        ftpdumper = self.get_client_for_url(urlremotefile)
        remotefile = self.get_remote_file(urlremotefile)
        return ftpdumper.download(remotefile, localfile)


class HTTPDumper(BaseDumper):
    """Dumper using HTTP protocol and "requests" library"""

    VERIFY_CERT = True
    IGNORE_HTTP_CODE = []  # list of HTTP code to ignore in case on non-200 response
    RESOLVE_FILENAME = False  # global trigger to get filenames from headers

    # when available

    def prepare_client(self):
        self.client = requests.Session()
        self.client.verify = self.__class__.VERIFY_CERT

    def need_prepare(self):
        return not self.client

    def release_client(self):
        self.client.close()
        self.client = None

    def remote_is_better(self, remotefile, localfile):
        """
        Determine if remote is better

        Override if necessary.
        """
        return True

    def download(self, remoteurl, localfile, headers={}):  # noqa: B006
        self.prepare_local_folders(localfile)
        res = self.client.get(remoteurl, stream=True, headers=headers)
        if not res.status_code == 200:
            if res.status_code in self.__class__.IGNORE_HTTP_CODE:
                self.logger.info("Remote URL %s gave http code %s, ignored" % (remoteurl, res.status_code))
                return
            else:
                raise DumperException(
                    "Error while downloading '%s' (status: %s, reason: %s)" % (remoteurl, res.status_code, res.reason)
                )
        # issue biothings.api #3: take filename from header if specified
        # note: this has to explicit, either on a globa (class) level or per file to dump
        if self.__class__.RESOLVE_FILENAME and res.headers.get("content-disposition"):
            parsed = cgi.parse_header(res.headers["content-disposition"])
            # looks like: ('attachment', {'filename': 'the_filename.txt'})
            if parsed and parsed[0] == "attachment" and parsed[1].get("filename"):
                # localfile is an absolute path, replace last part
                localfile = os.path.join(os.path.dirname(localfile), parsed[1]["filename"])
        self.logger.debug("Downloading '%s' as '%s'" % (remoteurl, localfile))
        fout = open(localfile, "wb")
        for chunk in res.iter_content(chunk_size=512 * 1024):
            if chunk:
                fout.write(chunk)
        fout.close()
        return res


class LastModifiedHTTPDumper(HTTPDumper, LastModifiedBaseDumper):
    """Given a list of URLs, check Last-Modified header to see
    whether the file should be downloaded. Sub-class should only have
    to declare SRC_URLS. Optionally, another field name can be used instead
    of Last-Modified, but date format must follow RFC 2616. If that header
    doesn't exist, it will always download the data (bypass)
    The release is generated from the last file's Last-Modified in SRC_URLS, and
    formatted according to RELEASE_FORMAT.
    """

    LAST_MODIFIED = "Last-Modified"
    ETAG = "ETag"
    RELEASE_FORMAT = "%Y-%m-%d"
    RESOLVE_FILENAME = True  # resolve by default as this dumper is called

    # with a list of URLs only, without any information
    # about the local filename to store data in

    def remote_is_better(self, remotefile, localfile):
        res = self.client.head(remotefile, allow_redirects=True)
        if self.__class__.LAST_MODIFIED not in res.headers:
            self.logger.warning(
                "Header '%s' doesn't exist, can determine if remote is better, assuming it is..."
                % self.__class__.LAST_MODIFIED
            )
            return True
        # In accordance with RFC 7231
        # The reason we are not using strptime is that it's locale sensitive
        # and changing locale and then changing it back is not thread safe.
        dt_tuple = email.utils.parsedate(res.headers[self.LAST_MODIFIED])
        # this utility function supports more malformed data so using this one
        if dt_tuple[5] == 60:
            _ = list(dt_tuple)
            _[5] = 59
            dt_tuple = tuple(_)
        # deal with potential leap second as defined in the RFC, good enough solution
        dt = datetime(*dt_tuple[:6], tzinfo=timezone.utc)  # HTTP-date is always in UTC
        remote_lastmodified = dt.timestamp()
        try:
            res = os.stat(localfile)
            local_lastmodified = int(res.st_mtime)
        except (FileNotFoundError, TypeError):
            return True  # doesn't even exist, need to dump
        if remote_lastmodified > local_lastmodified:
            self.logger.debug(
                "Remote file '%s' is newer (remote: %s, local: %s)"
                % (remotefile, remote_lastmodified, local_lastmodified)
            )
            return True
        else:
            return False

    def set_release(self):
        url = self.__class__.SRC_URLS[-1]
        res = self.client.head(url, allow_redirects=True)
        for _ in self.__class__.LAST_MODIFIED:
            try:
                remote_dt = datetime.strptime(res.headers[self.__class__.LAST_MODIFIED], "%a, %d %b %Y %H:%M:%S GMT")
                # also set release attr
                self.release = remote_dt.strftime(self.__class__.RELEASE_FORMAT)
            except KeyError:
                # Use entity tag (ETag) as version number. Remove weak ETag prefix.
                # Nginx marks an ETag as weak whenever a response body has been modified (including compression with gzip).
                # See: https://stackoverflow.com/questions/55305687/how-to-address-weak-etags-conversion-by-nginx-on-gzip-compression
                etag = res.headers[self.__class__.ETAG]
                if etag.startswith("W/"):
                    etag = etag[2:]
                self.release = etag.replace('"', "")


class WgetDumper(BaseDumper):
    def create_todump_list(self, force=False, **kwargs):
        """Fill self.to_dump list with dict("remote":remote_path,"local":local_path)
        elements. This is the todo list for the dumper. It's a good place to
        check whether needs to be downloaded. If 'force' is True though, all files
        will be considered for download"""
        raise NotImplementedError("Define in subclass")

    def prepare_client(self):
        """Check if 'wget' executable exists"""
        ret = os.system("type wget 2>&1 > /dev/null")
        if not ret == 0:
            raise DumperException("Can't find wget executable")

    def need_prepare(self):
        return False

    def release_client(self):
        pass

    def remote_is_better(self, remotefile, localfile):
        return True

    def download(self, remoteurl, localfile):
        self.prepare_local_folders(localfile)
        cmdline = "wget %s -O %s" % (remoteurl, localfile)
        return_code = os.system(cmdline)
        if return_code == 0:
            self.logger.info("Success.")
        else:
            self.logger.error("Failed with return code (%s)." % return_code)


class FilesystemDumper(BaseDumper):
    """
    This dumpers works locally and copy (or move) files to datasource folder
    """

    FS_OP = "cp"  # or 'mv' if file needs to be delete from original folder,

    # or 'ln' is a symlink should be created

    def prepare_client(self):
        """Check if 'cp' and 'mv' executable exists..."""
        for cmd in ["cp", "mv", "ln"]:
            ret = os.system("type %s 2>&1 > /dev/null" % cmd)
            if not ret == 0:
                raise DumperException("Can't find '%s' executable" % cmd)

    def need_prepare(self):
        return False

    def release_client(self):
        pass

    def remote_is_better(self, remotefile, localfile):
        res = os.stat(remotefile)
        remote_lastmodified = int(res.st_mtime)
        res = os.stat(localfile)
        local_lastmodified = int(res.st_mtime)
        if remote_lastmodified > local_lastmodified:
            return True
        else:
            return False

    def download(self, remotefile, localfile):
        self.prepare_local_folders(localfile)
        if self.__class__.FS_OP == "ln":
            cmdline = "rm -f %s && ln -s %s %s" % (localfile, remotefile, localfile)
        else:
            cmdline = "%s -f %s %s" % (self.__class__.FS_OP, remotefile, localfile)
        return_code = os.system(cmdline)
        if return_code == 0:
            self.logger.info("Success.")
        else:
            self.logger.error("Failed with return code (%s)." % return_code)


class DummyDumper(BaseDumper):
    """DummyDumper will do nothing...
    (useful for datasources that can't be downloaded anymore
    but still need to be integrated, ie. fill src_dump, etc...)
    """

    def __init__(self, *args, **kwargs):
        # make sure we don't create empty directory each time it's launched
        # so create a non-archiving dumper
        super(DummyDumper, self).__init__(archive=False, *args, **kwargs)
        self.release = ""

    def prepare_client(self):
        self.logger.info("Dummy dumper, will do nothing")
        pass

    async def dump(self, force=False, job_manager=None, *args, **kwargs):
        self.logger.debug("Dummy dumper, nothing to download...")
        self.prepare_local_folders(os.path.join(self.new_data_folder, "dummy_file"))
        # this is the only interesting thing happening here
        pinfo = self.get_pinfo()
        pinfo["step"] = "post_dump"
        job = await job_manager.defer_to_thread(pinfo, partial(self.post_dump, job_manager=job_manager))
        await asyncio.gather(job)  # consume future
        self.logger.info("Registering success")
        self.register_status("success")
        if self.__class__.AUTO_UPLOAD:
            set_pending_to_upload(self.src_name)
        self.logger.info("success", extra={"notify": True})


class ManualDumper(BaseDumper):
    """
    This dumper will assist user to dump a resource. It will usually expect the files
    to be downloaded first (sometimes there's no easy way to automate this process).
    Once downloaded, a call to dump() will make sure everything is fine in terms of
    files and metadata
    """

    def __init__(self, *args, **kwargs):
        super(ManualDumper, self).__init__(*args, **kwargs)
        # overide @property, it'll be set manually in this case (ie. not dynamically generated)
        # because it's a manual dumper and user specifies data folder path explicitely
        # (and see below)
        self._new_data_folder = None

    @property
    def new_data_folder(self):
        return self._new_data_folder

    @new_data_folder.setter
    def new_data_folder(self, value):
        self._new_data_folder = value

    def prepare(self, state={}):  # noqa : B006
        self.setup_log()
        if self.prepared:
            return
        if state:
            # let's be explicit, _state takes what it wants
            for k in self._state:
                self._state[k] = state[k]
            return
        self.prepare_src_dump()

    def prepare_client(self):
        self.logger.info("Manual dumper, assuming data will be downloaded manually")

    async def dump(self, path, release=None, force=False, job_manager=None, **kwargs):
        if os.path.isabs(path):
            self.new_data_folder = path
        elif path:
            self.new_data_folder = os.path.join(self.src_root_folder, path)
        else:
            self.new_data_folder = self.src_root_folder
        if release is None:
            # take latest path part, usually it's the release
            self.release = os.path.basename(self.new_data_folder)
        else:
            self.release = release
        # sanity check
        if not os.path.exists(self.new_data_folder):
            raise DumperException("Can't find folder '%s' (did you download data first ?)" % self.new_data_folder)
        if not os.listdir(self.new_data_folder):
            raise DumperException("Directory '%s' is empty (did you download data first ?)" % self.new_data_folder)

        pinfo = self.get_pinfo()
        pinfo["step"] = "post_dump"
        strargs = "[path=%s,release=%s]" % (self.new_data_folder, self.release)
        job = await job_manager.defer_to_thread(pinfo, partial(self.post_dump, job_manager=job_manager))
        await asyncio.gather(job)  # consume future
        # ok, good to go
        self.register_status("success")
        if self.__class__.AUTO_UPLOAD:
            set_pending_to_upload(self.src_name)
        self.logger.info("success %s" % strargs, extra={"notify": True})
        self.logger.info("Manually dumped resource (data_folder: '%s')" % self.new_data_folder)


class GoogleDriveDumper(HTTPDumper):
    def prepare_client(self):
        # FIXME: this is not very useful...
        super(GoogleDriveDumper, self).prepare_client()

    def remote_is_better(self, remotefile, localfile):
        return True

    def get_document_id(self, url):
        pr = urlparse.urlparse(url)
        if "drive.google.com/open" in url or "docs.google.com/uc" in url:
            q = urlparse.parse_qs(pr.query)
            doc_id = q.get("id")
            if not doc_id:
                raise DumperException("Can't extract document ID from URL '%s'" % url)
            return doc_id.pop()
        elif "drive.google.com/file" in url:
            frags = pr.path.split("/")
            ends = ["view", "edit"]
            if frags[-1] in ends:
                doc_id = frags[-2]
                return doc_id
            else:
                raise DumperException("URL '%s' doesn't end with %s, can't extract document ID" % (url, ends))

        raise DumperException("Don't know how to extract document ID from URL '%s'" % url)

    def download(self, remoteurl, localfile):
        """
        remoteurl is a google drive link containing a document ID, such as:
            - https://drive.google.com/open?id=<1234567890ABCDEF>
            - https://drive.google.com/file/d/<1234567890ABCDEF>/view

        It can also be just a document ID
        """
        try:
            from bs4 import BeautifulSoup
        except ImportError:
            raise ImportError('"BeautifulSoup4" is required for GoogleDriveDumper class: pip install beautifulsoup4')
        self.prepare_local_folders(localfile)
        if remoteurl.startswith("http"):
            doc_id = self.get_document_id(remoteurl)
        else:
            doc_id = remoteurl
        self.logger.info("Found document ID: %s" % doc_id)
        # first pass: get download URL with "confirm" code
        dl_url = "https://docs.google.com/uc?id=%s" % doc_id
        res = requests.get(dl_url)
        html = BeautifulSoup(res.text, "html.parser")
        link = html.find("a", {"id": "uc-download-link"})
        if not link:
            raise DumperException("Can't find a download link from '%s': %s" % (dl_url, html))
        href = link.get("href")
        # now build the final GET request, using cookies to simulate browser
        return super(GoogleDriveDumper, self).download(
            "https://docs.google.com" + href,
            localfile,
            headers={"cookie": res.headers["set-cookie"]},
        )


class GitDumper(BaseDumper):
    """
    Git dumper gets data from a git repo. Repo is stored in SRC_ROOT_FOLDER
    (without versioning) and then versions/releases are fetched in
    SRC_ROOT_FOLDER/<release>
    """

    GIT_REPO_URL = None
    DEFAULT_BRANCH = None

    def _get_remote_default_branch(self) -> Optional[bytes]:
        # expect bytes to work when invoking commands via subprocess
        # git doesn't really care about the encoding of refnames, it only
        # seems to be limited by the underlying filesystem
        # (don't count on the above statement, it's just an educated guess)
        cmd = ["git", "ls-remote", "--symref", self.GIT_REPO_URL, "HEAD"]
        try:
            # set locale to C so the output may have more reliable format
            result = subprocess.run(cmd, stdout=subprocess.PIPE, timeout=5, check=True, env={"LC_ALL": "C"})  # nosec
            r = re.compile(rb"^ref:\s+refs\/heads\/(.*)\s+HEAD$", flags=re.MULTILINE)
            m = r.match(result.stdout)
            if m is not None:
                return m[1]
        except (TimeoutError, subprocess.CalledProcessError):
            pass
        return None

    def _get_remote_branches(self) -> List[bytes]:
        cmd = ["git", "ls-remote", "--heads", self.GIT_REPO_URL]
        ret = []
        try:
            # set locale to C so the output may have more reliable format
            result = subprocess.run(cmd, stdout=subprocess.PIPE, timeout=5, check=True, env={"LC_ALL": "C"})  # nosec
            # user controls the URL anyways, and we don't use a shell
            # so it is safe
            r = re.compile(rb"^[0-9a-f]{40}\s+refs\/heads\/(.*)$", flags=re.MULTILINE)
            for m in re.findall(r, result.stdout):
                ret.append(m)
        except (TimeoutError, subprocess.CalledProcessError):
            pass
        return ret

    def _get_default_branch(self) -> Union[bytes, str]:
        # pylint: disable=bare-except
        # TODO: check TODO in _pull regarding original exclusive use of
        #  class variable
        # Case 1, default is explicitly set
        if self.DEFAULT_BRANCH is not None:
            return self.DEFAULT_BRANCH
        try:
            # Case 2, remote HEAD exists
            branch = self._get_remote_default_branch()
            if branch is not None:
                return branch
            # Case 3, 'main' exists but not 'master'
            branches = self._get_remote_branches()
            if b"main" in branches and b"master" not in branches:
                return "main"
        except:  # nosec  # noqa
            # fallback anything goes wrong
            pass
        # Case 4, use 'master' for compatibility reasons
        return "master"

    def _clone(self, repourl, localdir):
        self.logger.info("git clone '%s' into '%s'" % (repourl, localdir))
        subprocess.check_call(["git", "clone", repourl, localdir])

    def _pull(self, localdir, commit):
        # fetch+merge
        self.logger.info("git pull data (commit %s) into '%s'" % (commit, localdir))
        old = os.path.abspath(os.curdir)
        try:
            os.chdir(localdir)
            # discard changes, we don't want to activate a conflit resolution session...
            cmd = ["git", "reset", "--hard", "HEAD"]
            subprocess.check_call(cmd)
            # then fetch latest code (local repo, not applied to code base yet)
            cmd = ["git", "fetch", "--all"]
            subprocess.check_call(cmd)
            if commit != "HEAD":
                # first get the latest code from repo
                # (if a newly created branch is avail in remote, we can't check it out)
                self.logger.info("git checkout to commit %s" % commit)
                cmd = ["git", "checkout", commit]
                subprocess.check_call(cmd)
            else:
                # if we were on a detached branch (due to specific commit checkout)
                # we need to make sure to go back to master (re-attach)
                # TODO: figure out why it was originally using the class
                #  variable exclusively. Changed to prefer instance varaibles.
                branch = self._get_default_branch()
                cmd = ["git", "checkout", branch]
                subprocess.check_call(cmd)
            # then merge all remote changes to the current branch
            cmd = ["git", "merge"]
            subprocess.check_call(cmd)
            # and then get the commit hash
            out = subprocess.check_output(["git", "rev-parse", "--short", "HEAD"])
            self.release = f"{commit} {out.decode().strip()}"
        finally:
            os.chdir(old)
        pass

    @property
    def new_data_folder(self):
        # we don't keep release in data folder path
        # as it's a git repo
        return self.src_root_folder

    async def dump(self, release="HEAD", force=False, job_manager=None, **kwargs):
        assert self.__class__.GIT_REPO_URL, "GIT_REPO_URL is not defined"
        # assert self.__class__.ARCHIVE == False, "Git dumper can't keep multiple versions (but can move to a specific commit hash)"
        got_error = None
        self.release = release

        def do():
            do_clone = False
            if force:
                # force is also a way to clean and start from scratch
                rmdashfr(self.src_root_folder)
            if not os.path.exists(self.src_root_folder):
                # data folder doesn't even exist, no git files yet, we need to clone
                os.makedirs(self.src_root_folder)
                do_clone = True
            self.register_status("downloading", transient=True)
            if do_clone:
                self._clone(self.__class__.GIT_REPO_URL, self.src_root_folder)
            self._pull(self.src_root_folder, release)

        pinfo = self.get_pinfo()
        job = await job_manager.defer_to_thread(pinfo, partial(do))

        def done(f):
            nonlocal got_error
            try:
                _ = f.result()
                self.register_status("success")
            except Exception as e:
                got_error = e
                self.logger.exception("failed: %s" % e, extra={"notify": True})
                self.register_status("failed", download={"err": str(e)})
                raise

        job.add_done_callback(done)
        await job

    def prepare_client(self):
        """Check if 'git' executable exists"""
        ret = os.system("type git 2>&1 > /dev/null")
        if not ret == 0:
            raise DumperException("Can't find 'git' executable")

    def need_prepare(self):
        return True

    def release_client(self):
        pass

    def remote_is_better(self, remotefile, localfile):
        return True

    def download(self, remotefile, localfile):
        self.prepare_local_folders(localfile)
        cmdline = "wget %s -O %s" % (remotefile, localfile)
        return_code = os.system(cmdline)
        if return_code == 0:
            self.logger.info("Success.")
        else:
            self.logger.error("Failed with return code (%s)." % return_code)


####################


class DumperManager(BaseSourceManager):
    SOURCE_CLASS = BaseDumper

    def get_source_ids(self):
        """Return displayable list of registered source names (not private)"""
        # skip private ones starting with __
        # skip those deriving from bt.h.autoupdate.dumper.BiothingsDumper, they're used for autohub
        # and considered internal (note: only one dumper per source, so [0])
        from biothings.hub.autoupdate.dumper import BiothingsDumper

        registered = sorted(
            [
                src
                for src, klasses in self.register.items()
                if not src.startswith("__") and not issubclass(klasses[0], BiothingsDumper)
            ]
        )
        return registered

    def __repr__(self):
        return "<%s [%d registered]: %s>" % (
            self.__class__.__name__,
            len(self.register),
            self.get_source_ids(),
        )

    def clean_stale_status(self):
        # not uysing mongo query capabilities as hub backend could be ES, SQLlite, etc...
        # so manually iterate
        src_dump = get_src_dump()
        srcs = src_dump.find()
        for src in srcs:
            if src.get("download", {}).get("status", None) == "downloading":
                logging.warning("Found stale datasource '%s', marking download status as 'canceled'" % src["_id"])
                src["download"]["status"] = "canceled"
                src_dump.replace_one({"_id": src["_id"]}, src)

    def create_instance(self, klass):
        logging.debug("Creating new %s instance" % klass.__name__)
        inst = klass()
        return inst

    def register_classes(self, klasses):
        for klass in klasses:
            # supersede/monkey-patch klass with potiential existing conf values from DB
            btconfig.supersede(klass)
            if klass.SRC_NAME:
                if len(self.register.get(klass.SRC_NAME, [])) >= 1:
                    raise ResourceError(
                        "Can't register %s for source '%s', dumper already registered: %s"
                        % (klass, klass.SRC_NAME, self.register[klass.SRC_NAME])
                    )
                self.register.setdefault(klass.SRC_NAME, []).append(klass)
            else:
                try:
                    self.register[klass.name] = klass
                except AttributeError as e:
                    logging.error("Can't register class %s: %s" % (klass, e))
                    continue

    def dump_all(self, force=False, **kwargs):
        """
        Run all dumpers, except manual ones
        """
        jobs = []
        for src in self.register:
            job = self.dump_src(src, force=force, skip_manual=True, **kwargs)
            jobs.extend(job)
        return asyncio.gather(*jobs)

    def dump_src(self, src, force=False, skip_manual=False, schedule=False, check_only=False, **kwargs):
        if src in self.register:
            klasses = self.register[src]
        else:
            raise DumperException("Can't find '%s' in registered sources (whether as main or sub-source)" % src)

        jobs = []
        try:
            for _, klass in enumerate(klasses):
                if issubclass(klass, ManualDumper) and skip_manual:
                    logging.warning("Skip %s, it's a manual dumper" % klass)
                    continue
                crontab = None
                if schedule:
                    if klass.SCHEDULE:
                        crontab = klass.SCHEDULE
                    else:
                        raise DumperException("Missing scheduling information")
                job = self.job_manager.submit(
                    partial(
                        self.create_and_dump,
                        klass,
                        force=force,
                        job_manager=self.job_manager,
                        check_only=check_only,
                        **kwargs,
                    ),
                    schedule=crontab,
                )
                jobs.append(job)
            return jobs
        except Exception as e:
            logging.error("Error while dumping '%s': %s" % (src, e))
            raise

    def mark_success(self, src, dry_run=True):
        result = []
        if src in self.register:
            klasses = self.register[src]
        else:
            raise DumperException("Can't find '%s' in registered sources (whether as main or sub-source)" % src)
        for _, klass in enumerate(klasses):
            inst = self.create_instance(klass)
            result.append(inst.mark_success(dry_run=dry_run))
        return result

    def call(self, src, method_name, *args, **kwargs):
        """
        Create a dumper for datasource "src" and call method "method_name" on it,
        with given arguments. Used to create arbitrary calls on a dumper.
        "method_name" within dumper definition must a coroutine.
        """
        if src in self.register:
            klasses = self.register[src]
        else:
            raise DumperException("Can't find '%s' in registered sources (whether as main or sub-source)" % src)

        jobs = []
        try:
            for _, klass in enumerate(klasses):
                pfunc = partial(self.create_and_call, klass, method_name, *args, **kwargs)
                job = asyncio.ensure_future(pfunc())
                jobs.append(job)
            return jobs
        except Exception as e:
            logging.error("Error while dumping '%s': %s" % (src, e))
            raise

    async def create_and_dump(self, klass, *args, **kwargs):
        inst = self.create_instance(klass)
        res = await inst.dump(*args, **kwargs)
        return res

    async def create_and_call(self, klass, method_name, *args, **kwargs):
        inst = self.create_instance(klass)
        res = await getattr(inst, method_name)(*args, **kwargs)
        return res

    def schedule_all(self, raise_on_error=False, **kwargs):
        """
        Run all dumpers, except manual ones
        """
        errors = {}
        for src in self.register:
            try:
                self.dump_src(src, skip_manual=True, schedule=True, **kwargs)
            except Exception as e:
                errors[src] = e
                if raise_on_error:
                    raise
        if errors:
            logging.warning("Found errors while scheduling:\n%s" % pprint.pformat(errors))
            return errors

    def get_schedule(self, dumper_name):
        """Return the corresponding schedule for dumper_name
        Example result:
        {
            "cron": "0 9 * * *",
            "strdelta": "15h:20m:33s",
        }
        """
        info = None
        for sch in self.job_manager.loop._scheduled:
            if not isinstance(sch, asyncio.TimerHandle):
                continue
            if sch._cancelled:
                continue
            if sch._callback and dumper_name in str(sch._callback):
                info = job_renderer.cron_and_strdelta_info(sch)
                break
        return info

    def source_info(self, source=None):
        src_dump = get_src_dump()
        src_ids = self.get_source_ids()
        if source:
            if source in src_ids:
                src_ids = [source]
            else:
                return None
        res = []
        for _id in src_ids:
            src = src_dump.find_one({"_id": _id}) or {}
            assert len(self.register[_id]) == 1, "Found more than one dumper for source '%s': %s" % (
                _id,
                self.register[_id],
            )
            dumper = self.register[_id][0]
            name = "%s.%s" % (inspect.getmodule(dumper).__name__, dumper.__name__)
            bases = [
                "%s.%s" % (inspect.getmodule(k).__name__, k.__name__) for k in dumper.__bases__ if inspect.getmodule(k)
            ]
            schedule = self.get_schedule(name)
            src.setdefault("download", {})
            src["download"]["dumper"] = {
                "name": name,
                "bases": bases,
                "schedule": schedule,
                "manual": issubclass(dumper, ManualDumper),
            }
            src["name"] = _id
            src["_id"] = _id
            res.append(src)
        if source:
            return res.pop()
        else:
            return res

    def dump_info(self):
        res = {}
        for name, klasses in self.register.items():
            res[name] = [klass.__name__ for klass in klasses]
        return res


class APIDumper(BaseDumper):
    """
    Dump data from APIs

    This will run API calls in a clean process and write its results in
    one or more NDJSON documents.

    Populate the static methods get_document and get_release in your
    subclass, along with other necessary bits common to all dumpers.

    For details on specific parts, read the docstring for individual
    methods.

    An example subclass implementation can be found in the unii data
    source for MyGene.info.
    """

    _CHECK_JOIN_TIMEOUT = 20
    _TARGET_BUFFER_SIZE = 2 << 13  # 16KiB

    def create_todump_list(self, force=False, **kwargs):
        """
        This gets called by method `dump`, to populate self.to_dump
        """
        self.to_dump = [{"remote": "remote", "local": "local"}]
        # TODO: we can have get_release in another process as well
        #  but I don't think it is worth it.
        self.release = self.get_release()

    def remote_is_better(self, remotefile, localfile):
        """
        If there is a simple method to check whether remote is better
        """
        return True

    def download(self, remotefile, localfile):
        """
        Runs helper function in new process to download data

        This is run in a new process by the do_dump coroutine of the
        parent class. Then this will spawn another process that actually
        does all the work. This method is mostly for setting up the
        environment, setting up the the process pool executor to
        correctly use spawn and using concurrent.futures to simply run
        tasks in the new process, and periodically check the status
        of the task.

        Explanation: because this is actually running inside a process
        that forked from a threaded process, the internal state is more
        or less corrupt/broken, see `man 2 fork` for details.
        More discussions are in Slack from some time in 2021 on why it
        has to be forked and why it is broken.

        Caveats: the existing job manager will not know how much memory
        the actual worker process is using.
        """
        if not (remotefile == "remote") and (localfile == "local"):
            raise RuntimeError("This method is not supposed to be" "called outside dump/do_dump")
        wd = os.path.abspath(os.path.realpath(self.new_data_folder))
        os.makedirs(wd, exist_ok=True)
        self.to_dump = []
        mp_context = multiprocessing.get_context("spawn")
        # specifying mp_context is Python 3.7+ only
        executor = ProcessPoolExecutor(
            max_workers=1,
            mp_context=mp_context,
        )

        f = executor.submit(
            _run_api_and_store_to_disk,
            fn=self.get_document,
            buffer_size=self._TARGET_BUFFER_SIZE,
            working_directory=wd,
        )
        # we can schedule shutdown of the executor right now
        # but it has bugs, see below
        self.logger.debug("run_api submitted to executor...")
        ex = None
        while True:
            try:
                _ = f.result(timeout=self._CHECK_JOIN_TIMEOUT)
                self.logger.info("run_api exited successfully")
                break
            except concurrent.futures.TimeoutError:
                self.logger.debug("run_api is still running...")
            except concurrent.futures.CancelledError:
                self.logger.error("run_api has been unexpectedly cancelled...")
            except Exception as e:
                self.logger.warning("run_api exited with exception: %s", e)
                ex = e
                break
        # we could have scheduled the shutdown right after submitting the task
        # but it has some bugs, see https://bugs.python.org/issue39104
        executor.shutdown(wait=True)
        self.logger.info("executor shutdown successfully")
        if ex is not None:
            raise ex

    @staticmethod
    def get_document() -> Generator[Tuple[str, Any], None, None]:
        """
        Get document from API source

        Populate this method to yield documents to be stored on disk. Every
        time you want to save something to disk, do this:
        >>> yield 'name_of_file.ndjson', {'stuff': 'you want saved'}
        While the type definition says Any is accepted, it has to be JSON
        serilizable, so basically Python dictionaries/lists with strings and
        numbers as the most basic elements.

        Later on in your uploader, you can treat the files as NDJSON documents,
        i.e. one JSON document per line.

        It is recommended that you only do the minimal necessary processing in
        this step.

        A default HTTP client is not provided so you get the flexibility of
        choosing your most favorite tool.

        This MUST be a static method or it cannot be properly serialized to
        run in a separate process.

        This method is expected to be blocking (synchronous). However, be sure
        to properly SET TIMEOUTS. You open the resources here in this function
        so you have to deal with properly checking/closing them. If the
        invoker forcefully stops this method, it will leave a mess behind,
        therefore we do not do that.

        You can do a 5 second timeout using the popular requests package by
        doing something like this:
        >>> import requests
        >>> r = requests.get('https://example.org', timeout=5.0)
        You can catch the exception or setup retries. If you cannot handle
        the situation, just raise exceptions or not catch them. APIDumper
        will handle it properly: documents are only saved when the entire
        method completes successfully.
        """
        raise NotImplementedError

    @staticmethod
    def get_release() -> str:
        """
        Get the string for the release information.

        This is run in the main process and thread so it must return quickly.
        This must be populated

        Returns:
            string representing the release.
        """
        raise NotImplementedError

    @property
    def client(self):
        # overides the parent class
        return None

    def prepare_client(self):
        # having the client in the main process is not a good idea anyways
        # for some MongoDB client related things, it does make sense
        # but just closing the connection is not enough to free it from memory
        # and eliminate its threads.
        # The best way to do it is to run spawn a new process and run the client
        # there. Or do some kind of IPC and have the client in one process only.
        raise RuntimeError("prepare_client method of APIDumper and its " "descendents must not be called")

    def release_client(self):
        # dump will always call this method so we have to allow it
        if inspect.stack()[1].function == "dump":
            return
        raise RuntimeError("release_client method of APIDumper and its " "descendents must not be called")

    def need_prepare(self):
        raise RuntimeError("need_prepare method of APIDumper and its " "descendents must not be called")


def _run_api_and_store_to_disk(
    fn: Callable[[], Iterable[Tuple[str, Any]]],
    buffer_size: int,
    working_directory: str,
) -> None:
    """
    Runs an API Callable and Store result as NDJSON

    This is a helper function used by APIDumper and is supposed to be
    run in a separate process.

    It is defined in the module so that it can be serialized. The
    arguments must also be serializable.

    Args:
        fn: Callable (function or static method) that takes no arguments
            or keyword arguments. Must return an Iterable which
            individual items must be tuples. The said tuples must
            contain two elements, the first is a string which is the name
            of the output file, and the second is the object to be saved.
            The object must be JSON serializable.
        buffer_size: target buffer size per file, in bytes. It will always
            be overrun, but will immediately be written to the disk if it
            does overrun. This is so that very large operations will not
            cause out-of-memory errors. It is not for optimizing disk IO
            performance.
        working_directory: absolute path of the working directory. Files
            will be written in the given directory.
    """
    pid = os.getpid()
    ppid = os.getppid()  # TODO: check parent process  # noqa: F841
    if not os.path.isabs(working_directory):
        raise ValueError(f"desired working_directory {working_directory}" f" is not absolute")
    os.chdir(working_directory)
    buffer: Dict[str, bytearray] = {}
    try:
        for filename, obj in fn():
            fn_byte_arr = buffer.setdefault(filename, bytearray())
            fn_byte_arr.extend(orjson.dumps(obj) + b"\n")
            if len(fn_byte_arr) >= buffer_size:
                with open(f"{filename}.{pid}", "ab") as f:
                    f.write(fn_byte_arr)
                buffer[filename].clear()
    except Exception as e:
        # cleanup
        for filename in buffer.keys():
            if os.path.exists(filename):
                os.unlink(f"{filename}.{pid}")
        buffer.clear()
        raise e
    for filename, fn_byte_arr in buffer.items():
        with open(f"{filename}.{pid}", "ab") as f:
            f.write(fn_byte_arr)
    for filename in buffer.keys():
        os.rename(src=f"{filename}.{pid}", dst=filename)


class DockerContainerDumper(BaseDumper):
    """
    Start a docker container (typically runs on a different server) to prepare the data file on the remote container,
    and then download this file to the local data source folder.
    This dumper will do the following steps:
    - Booting up a container from provided parameters: image, tag, container_name.
    - The container entrypoint will be override by this long running command: "tail -f /dev/null"
    - When the container_name and image is provided together, the dumper will try to run the container_name.
        If the container with container_name does not exist, the dumper will start a new container from image param,
        and set its name as container_name.
    - Run the dump_command inside this container. This command MUST block the dumper until the data file is completely prepare.
        It will guarantees that the remote file is ready for downloading.
    - Run the get_version_cmd inside this container - if it provided. Set this command out put as self.release.
    - Download the remote file via Docker API, extract the downloaded .tar file.
    - When the downloading is complete:
        - if keep_container=false: Remove the above container after.
        - if keep_container=true: leave this container running.
    - If there are any execption when dump data, the remote container won't be removed, it will help us address the problem.


    These are supported connection types from the Hub server to the remote Docker host server:
        - ssh: Prerequisite: the SSH Key-Based Authentication is configured
        - unix: Local connection
        - http: Use an insecure HTTP connection over a TCP socket
        - https:  Use a secured HTTPS connection using TLS. Prerequisite:
            - The Docker API on the remote server MUST BE secured with TLS
            - A TLS key pair is generated on the Hub server and placed inside the same data plugin folder or the data source folder

    All info about Docker client connection MUST BE defined in the `config.py` file, under the DOCKER_CONFIG key, Ex
    Optional DOCKER_HOST can be used to override the docker connections in any docker dumper regardless of the value of the src_url
    For example, you can set DOCKER_HOST="localhost" for local testing:
        DOCKER_CONFIG = {
            "CONNECTION_NAME_1": {
                "tls_cert_path": "/path/to/cert.pem",
                "tls_key_path": "/path/to/key.pem",
                "client_url": "https://remote-docker-host:port"
            },
            "CONNECTION_NAME_2": {
                "client_url": "ssh://user@remote-docker-host"
            },
            "localhost": {
                "client_url": "unix://var/run/docker.sock"
            }
        }
        DOCKER_HOST = "localhost"

    The data_url should match the following format:
        docker://CONNECTION_NAME?image=DOCKER_IMAGE&tag=TAG&path=/path/to/remote_file&dump_command="this is custom command"&container_name=CONTAINER_NAME&keep_container=true&get_version_cmd="cmd"  # NOQA

    Supported params:
        - image: (Optional) the Docker image name
        - tag: (Optional) the image tag
        - container_name: (Optional) If this param is provided, the `image` param will be discard when dumper run.
        - path: (Required) path to the remote file inside the Docker container.
        - dump_command: (Required) This command will be run inside the Docker container in order to create the remote file.
        - keep_container: (Optional) accepted values: true/false, default: false.
            - If keep_container=true, the remote container will be persisted.
            - If keep_container=false, the remote container will be removed in the end of dump step.
        - get_version_cmd: (Optional) The custom command for checking release version of local and remote file. Note that:
            - This command must run-able in both local Hub (for checking local file) and remote container (for checking remote file).
            - "{}" MUST exists in the command, it will be replace by the data file path when dumper runs,
                ex: get_version_cmd="md5sum {} | awk '{ print $1 }'" will be run as: md5sum /path/to/remote_file | awk '{ print $1 }' and /path/to/local_file
    Ex:
      - docker://CONNECTION_NAME?image=IMAGE_NAME&tag=IMAGE_TAG&path=/path/to/remote_file(inside the container)&dump_command="run something with output is written to -O /path/to/remote_file (inside the container)"  # NOQA
      - docker://CONNECTION_NAME?container_name=CONTAINER_NAME&path=/path/to/remote_file(inside the container)&dump_command="run something with output is written to -O /path/to/remote_file (inside the container)&keep_container=true&get_version_cmd="md5sum {} | awk '{ print $1 }'"  # NOQA
      - docker://localhost?image=dockstore_dumper&path=/data/dockstore_crawled/data.ndjson&dump_command="/home/biothings/run-dockstore.sh"&keep_container=1
      - docker://localhost?image=dockstore_dumper&tag=latest&path=/data/dockstore_crawled/data.ndjson&dump_command="/home/biothings/run-dockstore.sh"&keep_container=True  # NOQA
      - docker://localhost?image=praqma/network-multitool&tag=latest&path=/tmp/annotations.zip&dump_command="/usr/bin/wget https://s3.pgkb.org/data/annotations.zip -O /tmp/annotations.zip"&keep_container=false&get_version_cmd="md5sum {} | awk '{ print $1 }'"  # NOQA
      - docker://localhost?container_name=<YOUR CONTAINER NAME>&path=/tmp/annotations.zip&dump_command="/usr/bin/wget https://s3.pgkb.org/data/annotations.zip -O /tmp/annotations.zip"&keep_container=true&get_version_cmd="md5sum {} | awk '{ print $1 }'"  # NOQA

    Container metadata:
    - All above params in the data_url can be pre-config in the Dockerfile by adding LABELs. This config will be used as the fallback of the data_url params:
        The dumper will find those params from both data_url and container metadata.
        If a param does not exist in data_url, dumper will use its value from container metadata (of course if it exist).
    For example:
        ... Dockerfile
        LABEL "path"="/tmp/annotations.zip"
        LABEL "dump_command"="/usr/bin/wget https://s3.pgkb.org/data/annotations.zip -O /tmp/annotations.zip"
        LABEL keep_container="true"
        LABEL desc=test
        LABEL container_name=mydocker
    """

    # the docker client connection name defined in the config file, default is localhost
    DOCKER_CLIENT_URL = None
    # attributes below are defined from either SRC_URLS or docker image LABEL metadata
    DOCKER_IMAGE = None  # should be in the format of <image_name>:<tag>
    CONTAINER_NAME = None
    DUMP_COMMAND = None
    GET_VERSION_CMD = None
    KEEP_CONTAINER = False
    DATA_PATH = None
    # set to 1 so we don't execute any docker command in parallel
    MAX_PARALLEL_DUMP = 1
    # default timeout to wait for container to start or stop
    # this does not impact the time to run dump_command or get_version_cmd
    # both commands will be run asynchronizely till they finish.
    TIMEOUT = 300
    # record the original container status if it exists before dumper runs
    # we will try to keep the container status as it was before dumper runs
    ORIGINAL_CONTAINER_STATUS = None

    def __init__(self, *args, **kwargs):
        if not docker_avail:
            raise ImportError('"docker" package is required for "DockerContainerDumper" class.')
        super().__init__(*args, **kwargs)
        # the following attributes are specific to DockerContainerDumper
        self.container = None
        self._source_info = {}  # parsed source_info dictionary from the SRC_URL string
        self.image_metadata = {}  # parsed LABEL metadata from the Docker image

    @property
    def source_config(self):
        if not self._source_info:
            url = self.__class__.SRC_URLS[0]
            self._source_info = docker_source_info_parser(url)
        return self._source_info

    def need_prepare(self):
        if not self.client:
            return True

    def prepare_client(self):
        self.logger.info("Preparing docker client...")
        if not isinstance(self.__class__.SRC_URLS, list):
            raise DumperException("SRC_URLS should be a list")
        elif not self.__class__.SRC_URLS:
            raise DumperException("SRC_URLS list is empty")
        elif len(self.__class__.SRC_URLS) > 1:
            raise DumperException("SRC_URLS list should contain only one source url string")
        if not self._state["client"]:
            if btconfig.DOCKER_HOST:
                docker_connection = btconfig.DOCKER_CONFIG.get(btconfig.DOCKER_HOST)
            else:
                docker_connection = btconfig.DOCKER_CONFIG.get(self.source_config["connection_name"])
            self.DOCKER_CLIENT_URL = docker_connection.get("client_url")
            parsed = urlparse.urlparse(docker_connection.get("client_url"))
            scheme = parsed.scheme
            self.logger.info(f"Prepare the connection to Docker server {self.DOCKER_CLIENT_URL}")
            use_ssh_client = False
            tls_config = None
            if scheme not in ["tcp", "unix", "http", "https", "ssh"]:
                raise DumperException(f"Connection scheme {scheme} is not supported")
            if scheme == "ssh":
                use_ssh_client = True
            if scheme == "https":
                cert_path = docker_connection.get("tls_cert_path")
                key_path = docker_connection.get("tls_key_path")
                if not cert_path or not key_path:
                    raise DumperException("Can not connect to the Docker server, missing cert info")
                tls_config = docker.tls.TLSConfig(client_cert=(cert_path, key_path), verify=False)

            if tls_config:
                client = docker.DockerClient(
                    base_url=self.DOCKER_CLIENT_URL, use_ssh_client=use_ssh_client, tls=tls_config
                )
            else:
                # when not using tls, we need to increase timeout for the client (1 week)
                client = docker.DockerClient(
                    base_url=self.DOCKER_CLIENT_URL, use_ssh_client=use_ssh_client, tls=tls_config, timeout=604800
                )
            if not client.ping():
                raise DumperException("Can not connect to the Docker server!")
            self.logger.info("Connected to Docker server")
            self._state["client"] = client
        # now prepare docker configs, which requires the docker client.
        self.prepare_dumper_params()
        return self._state["client"]

    def release_client(self):
        """Release the docker client connection, called in dump method"""
        if self.client:
            self.client.close()
            self.client = None
            self._state["client"] = None

    def set_dump_command(self):
        self.DUMP_COMMAND = self.source_config.get("dump_command") or self.image_metadata.get("dump_command")
        if not self.DUMP_COMMAND:
            raise DumperException('Missing the require "dump_command" parameter')

    def set_keep_container(self):
        self.KEEP_CONTAINER = (
            self.source_config["keep_container"]
            if "keep_container" in self.source_config
            else self.image_metadata.get("keep_container", "").lower() in {"true", "yes", "1", "y"}
        )

    def set_get_version_cmd(self):
        self.GET_VERSION_CMD = self.source_config.get("get_version_cmd") or self.image_metadata.get("get_version_cmd")

    def set_data_path(self):
        # TODO: we can see if we can support data path provided as a folder, instead a file. In this case, we should download the whole folder
        self.DATA_PATH = self.source_config.get("path") or self.image_metadata.get("path")
        if not self.DATA_PATH:
            raise DumperException('Missing the require "path" parameter')

    def get_remote_file(self):
        """return the remote file path within the container.
        In most of cases, dump_command should either generate this file or check if it's ready if there is another
        automated pipeline generates this file.
        """
        if not self.DATA_PATH:
            self.set_data_path()
        return self.DATA_PATH

    def prepare_dumper_params(self):
        """
        Read all docker dumper parameters from either the data plugin manifest or the Docker image or container metadata.
        Of course, at least one of docker_image or container_name parameters must be defined in the data plugin manifest first.
        If the parameter is not defined in the data plugin manifest, we will try to read it from the Docker image metadata.
        """
        self.DOCKER_IMAGE = self.source_config["docker_image"]
        self.CONTAINER_NAME = self.source_config["container_name"]
        if not self.DOCKER_IMAGE and not self.CONTAINER_NAME:
            raise DumperException("Either DOCKER_IMAGE or CONTAINER_NAME must be defined in the data plugin manifest")
        # now read the metadata from the Docker image or container, will check the image first, then the container
        if self.DOCKER_IMAGE:
            try:
                image = self.client.images.get(self.DOCKER_IMAGE)
                self.image_metadata = image.labels
            except ImageNotFound:
                raise DumperException(f'The docker image "{self.DOCKER_IMAGE}" is not found')
        elif self.CONTAINER_NAME:
            # when only container is provided, we will try to read the metadata from the container
            try:
                container = self.client.containers.get(self.CONTAINER_NAME)
                self.image_metadata = container.labels
            except NotFound:
                raise DumperException(f'The container "{self.CONTAINER_NAME}" is not found')
        else:
            self.image_metadata = {}

        # Fill dumper parameters if they aren't provided in the data plugin manifest.
        self.set_dump_command()
        self.set_keep_container()
        self.set_get_version_cmd()
        self.set_data_path()
        if not self.CONTAINER_NAME:
            # when the container_name is not provided in the data plugin manifest,
            # we will check image_metadata to see if it's defined there.
            self.CONTAINER_NAME = self.image_metadata.get("container_name")

    def prepare_remote_container(self):
        """prepare the remote container and set self.container, called in create_todump_list method"""
        if not self.container:
            try:
                self.container = self.client.containers.get(self.CONTAINER_NAME)
                self.ORIGINAL_CONTAINER_STATUS = self.container.status
            except (NotFound, NullResource):
                if self.DOCKER_IMAGE:
                    if self.CONTAINER_NAME:
                        self.logger.info(
                            'Can not find an existing container "%s", try to start a new one with this name.',
                            self.CONTAINER_NAME,
                        )
                        self.logger.info(
                            'Start a new container "%s" with a generic ENTRYPOINT: tail -f /dev/null',
                            self.CONTAINER_NAME,
                        )
                        self.container = self.client.containers.run(
                            image=self.DOCKER_IMAGE,
                            name=self.CONTAINER_NAME,
                            entrypoint="tail -f /dev/null",  # create a new container with a never end entrypoint
                            detach=True,
                            auto_remove=False,
                        )
                    else:
                        self.logger.info("Start a docker container with a generic ENTRYPOINT: tail -f /dev/null")
                        self.container = self.client.containers.run(
                            image=self.DOCKER_IMAGE,
                            entrypoint="tail -f /dev/null",  # create a new container with a never end entrypoint
                            detach=True,
                            auto_remove=False,
                        )
                        self.CONTAINER_NAME = self.container.name  # record the randomly generated container name

            except Exception as err:
                self.logger.exception(err)
                raise DockerContainerException("Docker APIError when trying to get the container")

            if self.container.status == "paused":
                self.logger.info('Unpause the container "%s" ...', self.CONTAINER_NAME)
                self.container.unpause()
                self.logger.info('The container "%s" is unpaused!', self.CONTAINER_NAME)
            elif self.container.status != "running":
                self.logger.info('Waiting for the container "%s"to run ...', self.CONTAINER_NAME)
                self.container.start()
                self.logger.info('The container "%s" is starting ...', self.CONTAINER_NAME)
            count = 0
            while self.container.status not in ["running", "exited"] and count < self.TIMEOUT:
                count += 1
                self.container.reload()  # Load this object from the server again and update attrs with the new data
                time.sleep(1)
            self.logger.info('The container "%s" is now running!', self.CONTAINER_NAME)
            self.container.reload()
            if self.container.status == "exited":
                raise DockerContainerException(f'The container "{self.CONTAINER_NAME}" exited')

    def set_release(self):
        """call the get_version_cmd to get the releases, called in get_todump_list method.
        if get_version_cmd is not defined, use timestamp as the release.

        This is currently a blocking method, assuming get_version_cmd is a quick command.
        But if necessary, we can make it async in the future.
        """
        _release = None
        if self.GET_VERSION_CMD:
            if self.container and self.container.status == "running":
                # we can't execute the cmd in a stop container
                exit_code, remote_version = self.container.exec_run(["sh", "-c", self.GET_VERSION_CMD])
                if exit_code == 0:
                    # make sure cmd run successfully on the remote
                    # the output of get_version_cmd should be the value returned from set_release method
                    _release = remote_version.decode().strip()
                else:
                    self.logger.error(
                        "Failed to run the provided get_version_cmd. Fallback to timestamp as the release."
                    )
            else:
                self.logger.error(
                    'The container "%s" is not running to run the provided get_version_cmd. Fallback to timestamp as the release.',
                    self.CONTAINER_NAME,
                )
        else:
            self.logger.info("get_version_cmd is not defined. Fallback to timestamp as the release.")
        # if _release is None, use timestamp as the release
        self.release = _release or datetime.now(timezone.utc).strftime("%Y-%m-%dT%H:%M:%SZ")

    def get_remote_lastmodified(self, remote_file):
        """get the last modified time of the remote file within the container using stat command"""
        if not self.CONTAINER_NAME:
            return None  # new fresh container, remote is always better
        try:
            container = self.client.containers.get(self.CONTAINER_NAME)
        except Exception:
            return None  # exception will be caught in the download step

        lastmodified = None
        if container.status == "running":
            docker_cmd = f'stat "{remote_file}" -c "%Y"'
            exit_code, remote_lastmodified_stdout = container.exec_run(["sh", "-c", docker_cmd])
            if exit_code == 0:
                lastmodified = int(remote_lastmodified_stdout.decode().strip())
            else:
                self.logger.error('Could not run "stat" on remote file in get_remote_lastmodified.')
        return lastmodified

    def remote_is_better(self, remote_file, local_file):
        try:
            res = os.stat(local_file)
        except (FileNotFoundError, TypeError):
            return True

        local_lastmodified = int(res.st_mtime)
        remote_lastmodified = self.get_remote_lastmodified(remote_file)
        if local_lastmodified != remote_lastmodified:
            return True
        return False

    def generate_remote_file(self):
        """Execute dump_command to generate the remote file, called in create_todump_list method"""
        if self.need_prepare():
            self.prepare_client()
        if self.DUMP_COMMAND:
            self.logger.info(f"Exec the command: sh -c {self.DUMP_COMMAND}")
            exit_code, output = self.container.exec_run(["sh", "-c", self.DUMP_COMMAND])
            self.logger.debug(output.decode())
            if exit_code != 0:
                remote_file = self.get_remote_file()
                self.logger.error(
                    'Failed to generate "%s", non-zero exit code from custom cmd: %s', remote_file, exit_code
                )
                self.logger.error(output)
                raise DumperException(f"Can not execute the command: {self.DUMP_COMMAND}")
        else:
            raise DockerContainerException("The dump_command parameter is must be defined")

    async def create_todump_list(self, force=False, job_manager=None, **kwargs):
        """
        Create the list of files to dump, called in dump method.
        This method will execute dump_command to generate the remote file in docker container,
        so we define this method as async to make it non-blocking.
        """
        if self.need_prepare():
            self.prepare_client()
        self.prepare_remote_container()
        # unprepare unpicklable objects so we can use multiprocessing
        state = self.unprepare()
        # set up job to generate remote file
        if job_manager:
            pinfo = self.get_pinfo()
            pinfo["step"] = "check"
            job = await job_manager.defer_to_process(pinfo, partial(self.generate_remote_file))
        else:
            # otherwise, just run it with asyncio loop directly
            async def run(fut):
                res = self.generate_remote_file()
                fut.set_result(res)

            loop = asyncio.get_event_loop()
            job = loop.create_future()
            loop.create_task(run(job))

        remote_error = False

        def remote_done(f):
            nonlocal remote_error
            remote_error = f.exception() or False

        job.add_done_callback(remote_done)
        await job
        if remote_error:
            raise remote_error
        # Need to reinit _state b/c of unprepare
        self.prepare(state)  # reverse of unpreare after async job is done
        # TODO: test if the following two lines can be removed after we call self.prepare(state) above
        if self.need_prepare():
            self.prepare_client()
        # self.setup_log()    # this line should not needed, since self.prepare calls it already.

        self.set_release()

        remote_file = self.get_remote_file()
        file_name = os.path.basename(remote_file)
        new_localfile = os.path.join(self.new_data_folder, file_name)
        remote_better = False
        if force:
            remote_better = True
        else:
            try:
                current_local_file = os.path.join(self.current_data_folder, file_name)
            except TypeError:
                # current data folder doesn't even exist
                current_local_file = new_localfile
            if current_local_file:
                remote_better = self.remote_is_better(remote_file, current_local_file)
            else:
                remote_better = True
        if remote_better:
            new_localfile = os.path.join(self.new_data_folder, file_name)
            self.to_dump.append({"remote": remote_file, "local": new_localfile})

    def prepare_local_folders(self, localfile):
        """prepare the local folder for the localfile, called in download method"""
        # the parent method will create the necessary folders
        super().prepare_local_folders(localfile)
        # remove the localfile if it exists
        if os.path.isfile(localfile):
            os.unlink(localfile)

    def download(self, remote_file, local_file):
        # removes local file if exists before downloading remote file to local
        self.prepare_local_folders(local_file)
        try:
            # get_archive returns a tar datastream and dict with stat info
            bits, stat = self.container.get_archive(remote_file, encode_stream=True)
            if stat.get("size", 0) > 0:
                tmp_file = f"{local_file}.tar"
                with open(tmp_file, "wb") as fp:
                    for chunk in bits:
                        fp.write(chunk)
                # untar data files from the tmp_file, the extracted file will have the same last modified date as the original remote file
                # so we don't need to synchronize the local file's lastmodified date.
                untarall(folder=os.path.dirname(local_file), pattern="*.tar")
                os.unlink(tmp_file)  # remove the temp tar file
                self.logger.info('Successfully downloaded file "%s" to "%s"', remote_file, local_file)
            else:
                raise DumperException('Find empty remote file: "{remote_file}"')
        except Exception as ex:
            self.logger.error('Failed to download remote file "%s" with exception: %s', remote_file, ex, exc_info=True)
            raise DumperException(f'Failed to download the remote file"{remote_file}"')

    def delete_or_restore_container(self):
        """Delete the container if it's created by the dumper, or restore it to its original status if it's pre-existing."""
        if self.container:
            if self.ORIGINAL_CONTAINER_STATUS:
                # if self.ORIGINAL_CONTAINER_STATUS is set, the container is pre-existing, we will not remove it
                # regardless of the keep_container parameter, and try to restore it to its original status.
                if not self.KEEP_CONTAINER:
                    self.logger.warning(
                        f'The container is pre-existing, "{self.container.name}", but "keep_container" is set to False. '
                        'We will ignore the "keep_container" setting, and won\'t remove the container.'
                    )
                if self.container.status == self.ORIGINAL_CONTAINER_STATUS:
                    # container is in the original status, do nothing
                    return
                elif self.ORIGINAL_CONTAINER_STATUS == "exited":
                    self.logger.debug(f'Stopping container: "{self.container.name}"')
                    self.container.stop()
                elif self.ORIGINAL_CONTAINER_STATUS == "paused":
                    self.logger.debug(f'Pausing container: "{self.container.name}"')
                    self.container.pause()
                else:
                    self.logger.warning(
                        f'The existing container status, "{self.container.status}", does not match its original status "{self.ORIGINAL_CONTAINER_STATUS}". '
                        "We will not touch it."
                    )
            elif not self.KEEP_CONTAINER:
                self.logger.debug(f"Removing container: {self.container.name}")
                self.container.stop()
                self.container.wait(timeout=self.TIMEOUT)
                self.container.remove(v=True)

    def post_dump(self, *args, **kwargs):
        """Delete container or restore the container status if necessary, called in the dump method after the dump is done (during the "post" step)"""
        super().post_dump(*args, **kwargs)
        self.delete_or_restore_container()



import logging
import xml.dom.minidom
from typing import NamedTuple
from xml.etree import ElementTree

from elasticsearch import Elasticsearch
from pymongo.collection import Collection


class _Ele(NamedTuple):  # Cleanup Element
    tag: str
    attrs: dict
    elems: list

    @classmethod
    def ment(cls, tag, attrs, content):  # _Ele.ment(..) :)
        return _Ele(tag, attrs, [_Ele.ment(*e) for e in content])

    def to_xml(self):
        attrs = self.attrs.copy()

        if self.tag in ("CleanUps", "Remove", "Keep"):
            attrs["size"] = str(len(self.elems))

        if self.tag == "Snapshot":
            attrs = {
                "_id": attrs["_id"],
                "build_name": attrs["build_name"],
                "created_at": str(attrs["created_at"]),
                "env": attrs.get("environment") or "N/A",
            }

        root = ElementTree.Element(self.tag, attrs)
        for elem in self.elems:
            root.append(elem.to_xml())
        return root

    def __str__(self):
        ets = ElementTree.tostring(self.to_xml())
        dom = xml.dom.minidom.parseString(ets)
        return dom.toprettyxml(indent=" " * 2)


def find(collection, env=None, keep=3, group_by=None, return_db_cols=False, **filters):
    if not isinstance(collection, Collection):
        raise NotImplementedError("Require MongoDB Hubdb.")

    if isinstance(group_by, (str, type(None))):
        group_by = "$" + (group_by or "build_config")
    elif isinstance(group_by, (list, tuple)):
        group_by = {k.replace(".", "_"): "$" + k for k in group_by}

    groups = list(
        collection.aggregate(
            [
                {"$project": {"build_config": "$build_config._id", "snapshot": {"$objectToArray": "$snapshot"}}},
                {"$unwind": {"path": "$snapshot"}},
                {
                    "$addFields": {
                        "snapshot.v.build_config": "$build_config",
                        "snapshot.v.build_name": "$_id",
                        "snapshot.v._id": "$snapshot.k",
                    }
                },
                {"$replaceRoot": {"newRoot": "$snapshot.v"}},
                {"$match": {"environment": env, **filters} if env else filters},
                {"$sort": {"created_at": 1}},
                {"$group": {"_id": group_by, "items": {"$push": "$$ROOT"}}},
            ]
        )
    )

    if return_db_cols:
        return groups

    return _Ele.ment(
        "CleanUps",
        {},
        [
            (
                "Group",
                _expand(group["_id"], group_by),
                [
                    ("Remove", {}, [("Snapshot", _doc, []) for _doc in _remove(group, keep)]),
                    ("Keep", {}, [("Snapshot", _doc, []) for _doc in _keep(group, keep)]),
                ],
            )
            for group in groups
        ],
    )


def _expand(group_id, group_by):
    if isinstance(group_id, str):
        return {group_by.strip("$"): group_id}
    if isinstance(group_id, dict):
        return group_id
    raise TypeError()


def _keep(doc, keep):
    return doc["items"][-keep or len(doc["items"]) :]


def _remove(doc, keep):
    return doc["items"][: -keep or len(doc["items"])]


# the operations below are not made async
# because SnapshotEnv.client is not async


def delete(collection, element, envs):
    cnt = 0
    assert element.tag == "CleanUps"
    for group in element.elems:
        for catagory in group.elems:
            if catagory.tag == "Remove":
                for snapshot in catagory.elems:
                    _delete(collection, snapshot, envs)
                    cnt += 1
    return cnt


def _delete(collection, snapshot, envs):
    assert snapshot.tag == "Snapshot"

    if "environment" in snapshot.attrs:
        env = snapshot.attrs["environment"]
        client = envs[env].client
    else:  # legacy format
        env = snapshot.attrs["conf"]["indexer"]["env"]
        env = envs.index_manager[env]
        client = Elasticsearch(**env["args"])

    client.snapshot.delete(
        snapshot.attrs["conf"]["repository"]["name"],
        snapshot.attrs["_id"],
    )

    collection.update_one(
        {"_id": snapshot.attrs["build_name"]},
        {"$unset": {f"snapshot.{snapshot.attrs['_id']}": 1}},
    )


def plain_text(element):
    plain_texts = []
    assert element.tag == "CleanUps"
    for group in element.elems:
        assert group.tag == "Group"

        plain_texts.append("Snapshots filtered by:")
        for k, v in group.attrs.items():
            plain_texts.append(f"      {k}={repr(v)}")
        plain_texts.append("")

        removes = group.elems[0].elems
        plain_texts.append(f"    Found {len(removes)} snapshots to remove:")
        for snapshot in removes:
            plain_texts.append(" " * 8 + _plain_text(snapshot))

        keeps = group.elems[1].elems
        plain_texts.append(f"    Found {len(keeps)} snapshots to keep:")
        for snapshot in keeps:
            plain_texts.append(" " * 8 + _plain_text(snapshot))
        plain_texts.append("")

    return "\n".join(plain_texts)


def _plain_text(snapshot):
    assert snapshot.tag == "Snapshot"
    return "".join(
        (
            snapshot.attrs["_id"],
            " (",
            f'env={snapshot.attrs.get("environment") or "N/A"}',
            ", "
            #
            # "build_name" generally agrees with the snapshot _id,
            # although technically snapshots can be named anything.
            # since in most use cases, the snapshot name at least
            # indicates which build it is created from, even when
            # it is not named exactly the same as the build, for
            # presentation concision, build_name is not shown here.
            # uncomment the following line if this assumption is
            # no longer true in the future.
            #
            # f'build_name={repr(snapshot.attrs["build_name"])}', ", ",
            f'created_at={str(snapshot.attrs["created_at"])}',
            ")",
        )
    )


# Feature Specification ↑
# https://suwulab.slack.com/archives/CC19LHAF2/p1631126588023700?thread_ts=1631063247.003700&cid=CC19LHAF2

# Snapshots filtered by:
#       build_config="demo_allspecies"
#        ...

#    Found 8 snapshots to remove:
#        ...
#    Found 3 snapshots to keep:
#        ...


def test_find():
    from pymongo import MongoClient

    logging.basicConfig(level="DEBUG")

    # mychem
    # -------
    # "su04"
    # "mychem_hubdb", "src_build"

    client = MongoClient("su06")
    collection = client["outbreak_hubdb"]["src_build"]

    print(plain_text(find(collection)))


def test_print():
    print(
        _Ele.ment(
            "A",
            {},
            [
                ("AA", {}, []),
                ("AB", {"ABC": "D"}, []),
            ],
        )
    )


if __name__ == "__main__":
    test_find()



from collections import defaultdict

from tornado.httpclient import AsyncHTTPClient
from tornado.web import RequestHandler

from biothings.web.analytics.channels import GA4Channel, GAChannel, SlackChannel


class Notifier:
    def __init__(self, settings):
        self.channels = []

        if hasattr(settings, "SLACK_WEBHOOKS"):
            self.channels.append(SlackChannel(getattr(settings, "SLACK_WEBHOOKS")))  # noqa B009
        if getattr(settings, "GA_ACCOUNT", None):
            self.channels.append(
                GAChannel(
                    getattr(settings, "GA_ACCOUNT"),  # noqa B009
                    getattr(settings, "GA_UID_GENERATOR_VERSION", 1),
                )
            )
        if getattr(settings, "GA4_MEASUREMENT_ID", None):
            self.channels.append(
                GA4Channel(
                    measurement_id=getattr(settings, "GA4_MEASUREMENT_ID"),  # noqa B009
                    api_secret=getattr(settings, "GA4_API_SECRET"),  # noqa B009
                    uid_version=getattr(settings, "GA4_UID_GENERATOR_VERSION", 2),
                )
            )

    def broadcast(self, event):
        for channel in self.channels:
            if channel.handles(event):
                yield from channel.send(event)


# Web Framework Support
# ----------------------------


# Tornado


# https://www.tornadoweb.org/en/stable/httputil.html
# #tornado.httputil.HTTPServerRequest.remote_ip
class AnalyticsMixin(RequestHandler):
    def on_finish(self):
        super().on_finish()

        if self.get_argument("no_tracking", False):
            return  # this feature is undocumented

        if self.settings.get("debug", False):
            return  # for testing and development

        # Make sure to start the server with xheaders=True so that
        # remote_ip considers X-Real-Ip and X-Forwarded-For headers

        request = defaultdict(type(None))
        request["user_ip"] = self.request.remote_ip
        request["user_agent"] = self.request.headers.get("User-Agent")
        request["host"] = self.request.host
        request["path"] = self.request.path
        request["referer"] = self.request.headers.get("Referer")
        self.event["__request__"] = request

        if hasattr(self, "biothings"):
            client = AsyncHTTPClient()
            notifier = self.biothings.notifier
            for request in notifier.broadcast(self.event):
                client.fetch(request)

        else:  # need to initialize a notifier
            raise NotImplementedError()


# FastAPI

# ...



import inspect
import logging
from pydoc import locate
from types import SimpleNamespace

from biothings.utils.info import DevInfo, FieldNote
from biothings.web import connections
from biothings.web.analytics.notifiers import Notifier
from biothings.web.options import OptionsManager as OptionSets
from biothings.web.query.builder import ESUserQuery, MongoQueryBuilder, SQLQueryBuilder
from biothings.web.query.engine import MongoQueryBackend, SQLQueryBackend
from biothings.web.query.formatter import MongoResultFormatter, SQLResultFormatter
from biothings.web.query.pipeline import MongoQueryPipeline, SQLQueryPipeline
from biothings.web.services.health import ESHealth, MongoHealth, SQLHealth
from biothings.web.services.metadata import BiothingsESMetadata, BiothingsMongoMetadata, BiothingsSQLMetadata

logger = logging.getLogger(__name__)


def load_class(kls):
    if inspect.isclass(kls):
        return kls
    if isinstance(kls, str):
        _kls = locate(kls)
        if _kls:  # could be None
            return _kls
    raise ValueError(kls)


def _requires(config_key):
    def requires(f):
        def _(self):
            if not getattr(self.config, config_key, None):
                return  # skip this database context
            return f(self)

        return _

    return requires


class BiothingsDBProxy:
    """
    Provide database-agnostic access to
    common biothings service components,
    for single database application.
    """

    def __init__(self):
        self.pipeline = None
        self.metadata = None
        self.health = None

    def configure(self, db):
        self.pipeline = db.pipeline
        self.metadata = db.metadata
        self.health = db.health


class BiothingsNamespace:
    # toolbox to use when writing handlers
    # contains common biothings services

    def __init__(self, config):
        self.config = config

        self.fieldnote = FieldNote(config.AVAILABLE_FIELDS_NOTES_PATH)
        self.devinfo = DevInfo()

        # web application
        self.notifier = Notifier(config)
        self.optionsets = OptionSets()
        self.handlers = {}

        # database access
        self.db = BiothingsDBProxy()
        self._configure_elasticsearch()
        self._configure_mongodb()
        self._configure_sql()

        # db uitility shortcuts
        self.pipeline = self.db.pipeline
        self.metadata = self.db.metadata
        self.health = self.db.health

    @_requires("ES_HOST")
    def _configure_elasticsearch(self):
        self.elasticsearch = SimpleNamespace()

        self.elasticsearch.client = connections.es.get_client(self.config.ES_HOST, **self.config.ES_ARGS)
        self.elasticsearch.async_client = connections.es.get_async_client(self.config.ES_HOST, **self.config.ES_ARGS)

        self.elasticsearch.userquery = ESUserQuery(self.config.USERQUERY_DIR)
        self.elasticsearch.metadata = BiothingsESMetadata(
            self.config.ES_INDICES,
            self.elasticsearch.async_client,
        )
        self.elasticsearch.pipeline = load_class(self.config.ES_QUERY_PIPELINE)(
            load_class(self.config.ES_QUERY_BUILDER)(
                self.elasticsearch.userquery,
                self.config.ANNOTATION_ID_REGEX_LIST,
                self.config.ANNOTATION_DEFAULT_SCOPES,
                self.config.ALLOW_RANDOM_QUERY,
                self.config.ALLOW_NESTED_AGGS,
                self.elasticsearch.metadata.biothing_metadata,
            ),
            load_class(self.config.ES_QUERY_BACKEND)(
                self.elasticsearch.async_client,
                self.config.ES_INDICES,
                self.config.ES_SCROLL_TIME,
                self.config.ES_SCROLL_SIZE,
            ),
            load_class(self.config.ES_RESULT_TRANSFORM)(
                self.elasticsearch.metadata.biothing_licenses,
                self.config.LICENSE_TRANSFORM,
                self.fieldnote.get_field_notes(),
                self.config.AVAILABLE_FIELDS_EXCLUDED,
            ),
            fetch_max_match=self.config.ANNOTATION_MAX_MATCH,
        )
        self.elasticsearch.health = ESHealth(self.elasticsearch.async_client, self.config.STATUS_CHECK)
        self.db.configure(self.elasticsearch)

    @_requires("MONGO_URI")
    def _configure_mongodb(self):
        self.mongo = SimpleNamespace()

        self.mongo.client = connections.mongo.get_client(self.config.MONGO_URI, **self.config.MONGO_ARGS)
        self.mongo.metadata = BiothingsMongoMetadata(
            self.config.MONGO_COLS,
            self.mongo.client,
        )
        self.mongo.pipeline = MongoQueryPipeline(
            MongoQueryBuilder(),
            MongoQueryBackend(self.mongo.client, self.config.MONGO_COLS),
            MongoResultFormatter(),
        )
        self.mongo.health = MongoHealth(self.mongo.client)
        self.db.configure(self.mongo)

    @_requires("SQL_URI")
    def _configure_sql(self):
        self.sql = SimpleNamespace()

        self.sql.client = connections.sql.get_client(self.config.SQL_URI, **self.config.SQL_ARGS)
        self.sql.metadata = BiothingsSQLMetadata(self.config.SQL_TBLS, self.sql.client)
        self.sql.pipeline = SQLQueryPipeline(
            SQLQueryBuilder(self.config.SQL_TBLS),
            SQLQueryBackend(self.sql.client),
            SQLResultFormatter(),
        )
        self.sql.health = SQLHealth(self.sql.client)
        self.db.configure(self.sql)



import logging
from collections import namedtuple
from enum import Enum
from functools import partial
from types import SimpleNamespace

from elasticsearch import Elasticsearch, helpers
from pymongo import MongoClient

from biothings.utils.es import ESIndex as BaseESIndex
from biothings.utils.loggers import get_logger

try:
    from biothings.utils.mongo import doc_feeder
except ImportError:
    import biothings

    biothings.config = SimpleNamespace()
    biothings.config.DATA_SRC_DATABASE = "biothings_src"
    biothings.config.DATA_TARGET_DATABASE = "biothings_build"
    from biothings.utils.mongo import doc_feeder

_IDExists = namedtuple("IDExists", ("id", "exists"))


class ESIndex(BaseESIndex):
    def __init__(self, client, index_name, **bulk_index_args):
        super().__init__(client, index_name)
        self.bulk_index_args = bulk_index_args

    # --------------------
    # bulk operations (m*)
    # --------------------

    def mget(self, ids):
        """Return a list of documents like
        [
            { "_id": "0", "a": "b" },
            { "_id": "1", "c": "d" },
            # 404s are skipped
        ]
        """
        response = self.client.mget(
            body={"ids": ids},
            index=self.index_name,
            doc_type=self.doc_type,
        )
        for doc in response["docs"]:
            if doc.get("found"):
                doc["_source"]["_id"] = doc["_id"]
                yield doc["_source"]

    def mexists(self, ids):
        """Return a list of tuples like
        [
            (_id_0, True),
            (_id_1, False),
            (_id_2, True),
            ....
        ]
        """
        res = self.client.search(
            index=self.index_name,
            doc_type=self.doc_type,
            body={"query": {"ids": {"values": ids}}},
            stored_fields=None,
            _source=None,
            size=len(ids),
        )
        id_set = {doc["_id"] for doc in res["hits"]["hits"]}
        return [_IDExists(_id, _id in id_set) for _id in ids]

    def mindex(self, docs):
        """Index and return the number of docs indexed."""

        def _action(doc):
            _doc = {
                "_index": self.index_name,
                "_type": self.doc_type,
                "_op_type": "index",
            }
            _doc.update(doc)  # with _id
            return _doc

        return helpers.bulk(self.client, map(_action, docs), **self.bulk_index_args)[0]

    # NOTE
    # Why doesn't "mget", "mexists", "mindex" belong to the base class?
    # At this moment, their interfaces are too opinionated/customized
    # for usage in this module. Unless we find them directly useful in
    # another module in the future, proving their genericity, they should
    # stay close in this module only.


# Data Collection Client


def _get_es_client(es_client_args, es_blk_args, es_idx_name):
    return ESIndex(Elasticsearch(**es_client_args), es_idx_name, **es_blk_args)


def _get_mg_client(mg_client_args, mg_dbs_name, mg_col_name):
    return MongoClient(**mg_client_args)[mg_dbs_name][mg_col_name]


# --------------
#  Entry Point
# --------------


def dispatch(
    mg_client_args,
    mg_dbs_name,
    mg_col_name,
    es_client_args,
    es_blk_args,
    es_idx_name,
    ids,
    mode,
    name,
):
    return IndexingTask(
        partial(_get_es_client, es_client_args, es_blk_args, es_idx_name),
        partial(_get_mg_client, mg_client_args, mg_dbs_name, mg_col_name),
        ids,
        mode,
        f"index_{es_idx_name}",
        name,
    ).dispatch()


def _ensure_logger(logger):
    if not logger:
        return logging.getLogger(__name__)
    if isinstance(logger, str):
        return get_logger(logger)[0]
    return logger


def _validate_ids(ids, logger=None):
    validated_ids = []
    invalid_ids = []
    for _id in ids:
        if not isinstance(_id, str):
            raise TypeError("_id '%s' has invalid type (!str)." % repr(_id))
        if len(_id) > 512:  # this is an ES limitation
            invalid_ids.append(_id)
            message = "_id is too long: '%s'" % _id
            if logger:
                logger.warning(message)
            else:
                print(message)
        else:
            validated_ids.append(_id)
    return validated_ids, invalid_ids


class Mode(Enum):
    INDEX = "index"
    PURGE = "purge"  # same as 'index' in this module
    MERGE = "merge"
    RESUME = "resume"


class IndexingTask:
    """
    Index one batch of documents from MongoDB to Elasticsearch.
    The documents to index are specified by their ids.
    """

    def __init__(self, es, mongo, ids, mode=None, logger=None, name="task"):
        assert callable(es)
        assert callable(mongo)

        self.logger = _ensure_logger(logger)
        self.name = f"#{name}" if isinstance(name, int) else name

        self.ids, self.invalid_ids = _validate_ids(ids, self.logger)
        self.mode = Mode(mode or "index")

        # these are functions to create clients,
        # each also associated with an organizational
        # structure in the corresponding database,
        # functioning as the source or destination
        # of batch document manipulation.
        self.backend = SimpleNamespace()
        self.backend.es = es  # wrt an index
        self.backend.mongo = mongo  # wrt a collection

    def _get_clients(self):
        clients = SimpleNamespace()
        clients.es = self.backend.es()
        clients.mongo = self.backend.mongo()
        return clients

    def dispatch(self):
        if self.mode in (Mode.INDEX, Mode.PURGE):
            return self.index()
        elif self.mode == Mode.MERGE:
            return self.merge()
        elif self.mode == Mode.RESUME:
            return self.resume()

    def index(self):
        clients = self._get_clients()
        docs = doc_feeder(
            clients.mongo,
            step=len(self.ids),
            inbatch=False,
            query={"_id": {"$in": self.ids}},
        )
        self.logger.info("%s: %d documents.", self.name, len(self.ids))
        count_docs = clients.es.mindex(docs)
        return count_docs + len(self.invalid_ids)

    def merge(self):
        clients = self._get_clients()

        upd_cnt, docs_old = 0, {}
        new_cnt, docs_new = 0, {}

        # populate docs_old
        for doc in clients.es.mget(self.ids):
            docs_old[doc["_id"]] = doc

        # populate docs_new
        for doc in doc_feeder(
            clients.mongo,
            step=len(self.ids),
            inbatch=False,
            query={"_id": {"$in": self.ids}},
        ):
            docs_new[doc["_id"]] = doc
            doc.pop("_timestamp", None)

        # merge existing ids
        for key in list(docs_new):
            if key in docs_old:
                docs_old[key].update(docs_new[key])
                del docs_new[key]

        # updated docs (those existing in col *and* index)
        upd_cnt = clients.es.mindex(docs_old.values())
        self.logger.info("%s: %d documents updated.", self.name, upd_cnt)

        # new docs (only in col, *not* in index)
        new_cnt = clients.es.mindex(docs_new.values())
        self.logger.info("%s: %d new documents.", self.name, new_cnt)

        return upd_cnt + new_cnt

    def resume(self):
        clients = self._get_clients()
        missing_ids = [x.id for x in clients.es.mexists(self.ids) if not x.exists]
        self.logger.info("%s: %d missing documents.", self.name, len(missing_ids))
        if missing_ids:
            self.ids = missing_ids
            self.index()
        return len(self.ids)


def test_00():  # ES
    from pprint import pprint as print

    index = ESIndex(Elasticsearch(), "mynews_202105261855_5ffxvchx")
    print(index.doc_type)
    print(
        list(
            index.mget(
                [
                    "0999b13cb8026aba",
                    "1111647aaf9c70b4",
                    "________________",
                ]
            )
        )
    )
    # print(list(index.mexists([
    #     "0999b13cb8026aba",
    #     "1111647aaf9c70b4",
    #     "________________"
    # ])))


def test_clients():
    def _mongo():
        client = MongoClient()
        database = client["biothings_build"]
        return database["mynews_202012280220_vsdevjdk"]

    def _es():
        client = Elasticsearch()
        return ESIndex(client, "indexer-test")

    return (_es, _mongo)


def test0():
    task = IndexingTask(
        *test_clients(),
        ("0999b13cb8026aba", "1111647aaf9c70b4", "1c9828073bad510c"),
    )
    task.index()


def test1():
    task = IndexingTask(
        *test_clients(),
        (
            "0999b13cb8026aba",
            "1111647aaf9c70b4",
            "1c9828073bad510c",
            "1f447d7fc6dcc2cf",
            "27e81a308e4e04da",
        ),
    )
    task.resume()


if __name__ == "__main__":
    logging.basicConfig(level="DEBUG")
    test_00()
    # test0()
    # test1()



from typing import NamedTuple


class _version_info(NamedTuple):
    # similar to sys._version_info
    major: int
    minor: int
    micro: int


version_info = _version_info(0, 12, "3")
__version__ = ".".join(map(str, version_info))


def config_for_app(config):
    import warnings

    warnings.warn(
        'It is safe to remove this function call now, just "import biothings.hub" will take care of it.',
        DeprecationWarning,
        stacklevel=2,
    )



import elasticsearch

from biothings.utils.exceptions import RepositoryVerificationFailed


class Repository:
    def __init__(self, client, repository):
        # Corresponds to HTTP operations on
        # /_snapshot/<repository>

        self.client = client
        self.name = repository

    def exists(self):
        try:
            self.client.snapshot.get_repository(self.name)
        except elasticsearch.exceptions.NotFoundError:
            return False
        return True

    def create(self, **body):
        # https://www.elastic.co/guide/en/elasticsearch/plugins/current/repository-s3-client.html
        return self.client.snapshot.create_repository(self.name, body=body)

    def delete(self):
        self.client.snapshot.delete_repository(self.name)

    def verify(self, config):
        """A repository is consider properly setup and working, when:
        - passes verification of ElasticSearch
        - it's settings must match with the snapshot's config.
        """

        # Check if the repo pass ElasticSearch's verification
        self.client.snapshot.verify_repository(self.name)

        # Check if the repo's settings match with the snapshot's config
        repo_settings = self.client.snapshot.get_repository(self.name)[self.name]
        incorrect_data = {}

        for field in ["type", "settings"]:
            if config[field] != repo_settings[field]:
                incorrect_data[field] = {"config": config[field], "repo": repo_settings[field]}

        if incorrect_data:
            raise RepositoryVerificationFailed(
                {
                    "error": "repository_verification_exception",
                    "detail": {
                        "message": "the repository's settings is not match with snapshot config.",
                        "diff": incorrect_data,
                    },
                }
            )

    def __str__(self):
        return (
            f"<Repository {'READY' if self.exists() else 'MISSING'}"
            f" name='{self.name}'"
            f" client={self.client}"
            f">"
        )


def test_01():
    from elasticsearch import Elasticsearch

    client = Elasticsearch()
    snapshot = Repository(client, "mynews")
    print(snapshot)


def test_02():
    from elasticsearch import Elasticsearch

    client = Elasticsearch()
    snapshot = Repository(client, "______")
    print(snapshot)


if __name__ == "__main__":
    test_01()



import asyncio
import logging
from collections import Counter
from dataclasses import dataclass

import elasticsearch

from biothings.web.query.builder import RawQueryInterrupt
from biothings.web.query.engine import EndScrollInterrupt, RawResultInterrupt

# here this module defines two types of operations supported in
# each query pipeline class, one called "search" which corresponds to
# a key-value pair search, with the sometimes optional parameter "scopes"
# being a list of keys to look for the value, named as parameter "q".
# the other type of query is called "fetch", that looks up documents
# basing on pre-defined fields that function as their mostly unique
# identifiers, it is a special type of "search" that ususally returns
# only one document. in this type of query, "scopes" are absolutely NOT
# provided, and the query term is called "id" in the parameter list.

# this module is called pipeline because
# a) it is a combination of individual query processing stages
# b) during async operations, it functions like the CPU pipeline,
#    more than 1 stage can be busy at a single point in time.

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class QueryPipelineException(Exception):
    code: int = 500
    summary: str = ""
    details: object = None
    # use code here to indicate error types instead of
    # exception types, this reduces the number of potential
    # exception types this module needs to create.
    # furthermore, consider using a superset of HTTP codes,
    # so that error translation from the upper layers
    # is also convenient and straightforward.


class QueryPipelineInterrupt(QueryPipelineException):
    def __init__(self, data):
        super().__init__(200, None, data)


class QueryPipeline:
    def __init__(self, builder, backend, formatter, **settings):
        self.builder = builder
        self.backend = backend
        self.formatter = formatter
        self.settings = settings

    def search(self, q, **options):
        query = self.builder.build(q, **options)
        result = self.backend.execute(query, **options)
        return self.formatter.transform(result, **options)

    def fetch(self, id, **options):
        assert options.get("scopes") is None
        result = self.search(id, **options)
        return result


def _simplify_ES_exception(exc, debug=False):
    result = {}
    try:
        root_cause = exc.info.get("error", exc.info)
        root_cause = root_cause["root_cause"][0]["reason"]
        root_cause = root_cause.replace('"', "'").split("\n")
        for index, cause in enumerate(root_cause):
            result["root_cuase_line_" + f"{index:02}"] = cause
    except IndexError:
        pass  # no root cause
    except Exception:
        logger.exception(
            " ".join(
                (
                    "Unexpected error in _simplify_ES_exception.",
                    "Caused by incompatible version, build, etc.",
                    "Update ES exception parsing logic here.",
                )
            )
        )

    if debug:  # raw ES error response
        result["debug"] = exc.info

    return exc.error, result


def capturesESExceptions(func):
    async def _(*args, **kwargs):
        try:
            return await func(*args, **kwargs)
        except (
            RawQueryInterrupt,  # correspond to 'rawquery' option
            RawResultInterrupt,  # correspond to 'raw' option
            EndScrollInterrupt,
        ) as exc:
            raise QueryPipelineInterrupt(exc.data)

        except AssertionError as exc:
            # in our application, AssertionError should be internal
            # the individual components raising the error should instead
            # rasie exceptions like ValueError and TypeError for bad input
            logging.exception("FIXME: Unexpected Assertion Error.", exc_info=exc)
            raise QueryPipelineException(500, str(exc) or "N/A")

        except (ValueError, TypeError) as exc:
            raise QueryPipelineException(400, type(exc).__name__, str(exc))

        except elasticsearch.ConnectionError:  # like timeouts..
            raise QueryPipelineException(503)

        except elasticsearch.RequestError as exc:  # 400s
            raise QueryPipelineException(400, *_simplify_ES_exception(exc))

        # it seems like the managed Elasticsearch service by AWS
        # may provide slightly different exceptions when the server
        # is overloaded comparing to that of self-managed ES.

        # this case and most of the handling below can be further studied.
        # most of the exception handlings from this point on are based on
        # experience. further documentation in details will be helpful.

        except elasticsearch.TransportError as exc:  # >400
            if exc.error == "search_phase_execution_exception":
                reason = exc.info.get("caused_by", {}).get("reason", "")

                if "rejected execution" in reason:
                    raise QueryPipelineException(503)
                else:  # unexpected, provide additional information for debug
                    raise QueryPipelineException(500, *_simplify_ES_exception(exc, True))

            elif exc.error == "index_not_found_exception":
                raise QueryPipelineException(500, exc.error)

            elif exc.status_code in (429, "N/A"):
                raise QueryPipelineException(503)
            else:  # unexpected
                raise

    return _


class AsyncESQueryPipeline(QueryPipeline):
    @capturesESExceptions
    async def search(self, q, **options):
        if isinstance(q, list):  # multisearch
            options["templates"] = (dict(query=_q) for _q in q)
            options["template_miss"] = dict(notfound=True)
            options["template_hit"] = dict()

        query = self.builder.build(q, **options)
        response = await self.backend.execute(query, **options)
        result = self.formatter.transform(response, **options)
        return result

    @capturesESExceptions
    async def fetch(self, id, **options):
        if options.get("scopes"):
            raise ValueError("Scopes Not Allowed.")

        # for builder
        options["autoscope"] = True
        # for formatter
        options["version"] = True
        options["score"] = False
        options["one"] = True

        # annotation endpoint should work on fields with reasonable
        # uniqueness of values, like id and symbol fields. because
        # we do not provide pagination for this endpoint, as it is
        # largely unncessary, in the case of matching too many docs
        # for one request, raise an exception instead of providing
        # the user incomplete matches. usually, when this happends,
        # it indicates bad choices of values as the default fields.

        MAX_MATCH = self.settings.get("fetch_max_match", 1000)
        options["size"] = MAX_MATCH + 1  # err when len(res) > MAX

        # "fetch" is a wrapper over "search".
        # ----------------------------------------
        result = await self.search(id, **options)
        # ----------------------------------------

        if isinstance(id, list):  # batch
            counter = Counter(x["query"] for x in result)
            if counter.most_common(1)[0][1] > MAX_MATCH:
                raise QueryPipelineException(500, "Too Many Matches.")

        else:  # single
            if result is None:
                raise QueryPipelineException(404, "Not Found.")

            if isinstance(result, list) and len(result) > MAX_MATCH:
                raise QueryPipelineException(500, "Too Many Matches.")

        return result


class ESQueryPipeline(QueryPipeline):  # over async client
    # These implementations may not be performance optimized
    # It is used as a proof of concept and helps the design
    # of upper layer constructs, it also simplifies testing
    # by providing ioloop management, enabling sync access.

    # >>> from biothings.web.query.pipeline import ESQueryPipeline
    # >>> pipeline = ESQueryPipeline()
    #
    # >>> pipeline.fetch("1017", _source=["symbol"])
    # {'_id': '1017', '_version': 1, 'symbol': 'CDK2'}
    #
    # >>> pipeline.search("1017", _source=["symbol"])
    # {
    #   'took': 11,
    #   'total': 1,
    #   'max_score': 4.0133753,
    #   'hits': [
    #       {
    #           '_id': '1017',
    #           '_score': 4.0133753,
    #           'symbol': 'CDK2'
    #       }
    #   ]
    # }

    def __init__(self, builder=None, backend=None, formatter=None, *args, **kwargs):
        if not builder:
            from biothings.web.query.builder import ESQueryBuilder

            builder = ESQueryBuilder()

        if not backend:
            from biothings.web.connections import get_es_client
            from biothings.web.query.engine import ESQueryBackend

            client = get_es_client(async_=True)
            backend = ESQueryBackend(client)

        if not formatter:
            from biothings.web.query.formatter import ESResultFormatter

            formatter = ESResultFormatter()

        super().__init__(builder, backend, formatter, *args, **kwargs)

    def _run_coroutine(self, coro, *args, **kwargs):
        loop = asyncio.get_event_loop()
        pipeline = AsyncESQueryPipeline(self.builder, self.backend, self.formatter)
        return loop.run_until_complete(coro(pipeline, *args, **kwargs))

    def search(self, q, **options):
        return self._run_coroutine(AsyncESQueryPipeline.search, q, **options)

    def fetch(self, id, **options):
        return self._run_coroutine(AsyncESQueryPipeline.fetch, id, **options)


class MongoQueryPipeline(QueryPipeline):
    pass


class SQLQueryPipeline(QueryPipeline):
    pass



"""A Programmatic Query API supporting Biothings Query Syntax.

From an architecture perspective, :py:mod:`biothings.web.query` is one of
the data services, built on top of the :py:mod:`biothings.web.connections`
layer, however, due to the complexity of the module, it is escalated one
level in organization to simplify the overall folder structure. The features
are available in `biothings.web.services.query` namespace via import.

"""

from biothings.web.query import *  # noqa F403 F401



"""
    Biothings Query Builder

    Turn the biothings query language to that of the database.
    The interface contains a query term (q) and query options.

    Depending on the underlying database choice, the data type
    of the query term and query options vary. At a minimum,
    a query builder should support:

    q: str, a query term,
        when not provided, always perform a match all query.
        when provided as an empty string, always match none.

    options: dotdict, optional query options.

        scopes: list[str], the fields to look for the query term.
                the meaning of scopes being an empty list or a
                None object/not provided is controlled by specific
                class implementations or not defined.

        _source: list[str], fields to return in the result.
        size: int, maximum number of hits to return.
        from_: int, starting index of result to return.
        sort: str, customized sort keys for result list

        aggs: str, customized aggregation string.
        post_filter: str, when provided, the search hits are filtered after the aggregations are calculated.
        facet_size: int, maximum number of agg results.

"""
import logging
import os
import re
from collections import UserString, namedtuple
from copy import deepcopy
from random import randrange

import orjson
from elasticsearch_dsl import MultiSearch, Q, Search
from elasticsearch_dsl.exceptions import IllegalOperation

from biothings.utils.common import dotdict


class RawQueryInterrupt(Exception):
    def __init__(self, data):
        super().__init__()
        self.data = data


Query = namedtuple("Query", ("term", "scopes"))
Group = namedtuple("Group", ("term", "scopes"))


class QStringParser:
    def __init__(
        self,
        default_scopes=("_id",),
        patterns=((r"(?P<scope>\w+):(?P<term>[^:]+)", ()),),
        gpnames=("term", "scope"),
    ):
        assert isinstance(default_scopes, (tuple, list))
        assert all(isinstance(field, str) for field in default_scopes)
        self.default = default_scopes  # ["_id", "entrezgene", "ensembl.gene"]
        self.patterns = []  # [(re.compile(r'^\d+$'), ['entrezgene', 'retired'])]
        self.gpname = Group(*gpnames)  # symbolic group name for term substitution
        for pattern, fields in patterns:
            fields = [fields] if isinstance(fields, str) else fields
            assert all(isinstance(field, str) for field in fields)
            pattern = re.compile(pattern) if isinstance(pattern, str) else pattern
            if hasattr(re, "Pattern"):  # TODO remove for python>3.7
                assert isinstance(pattern, re.Pattern)
            self.patterns.append((pattern, fields))

    def parse(self, q):
        assert isinstance(q, str)
        for regex, fields in self.patterns:
            match = re.fullmatch(regex, q)
            if match:
                named_groups = match.groupdict()
                q = named_groups.get(self.gpname.term) or q
                _fields = named_groups.get(self.gpname.scopes)
                fields = [_fields] if _fields else fields or self.default
                return Query(q, fields)
        return Query(q, self.default)


class ESScrollID(UserString):
    def __init__(self, seq: object):
        super().__init__(seq)
        # scroll id cannot be empty
        assert self.data


#
#             ES Query Builder Architecture
# -------------------------------------------------------
#                         build
#                 (support multisearch)
# --------------------------↓↓↓--------------------------
#                        _build_one
#  (dispatch basing on scopes, then apply_extras(..))
# ------------↓↓↓------------------------↓↓↓-------------
#    _build_string_query    |  _build_match_query
#  (__all__, userquery,..)  | (compound match query)
# ------------↓↓↓------------------------↓↓↓-------------
#    default_string_query   |   default_match_query
#  (map to ES query string) | (map to ES match query)
# -------------------------------------------------------


class ESQueryBuilder:
    """
    Build an Elasticsearch query with elasticsearch-dsl.
    """

    # Different from other query pipelines, elasticsearch
    # supports querystring query, which means we can directly
    # dispatch queires without fields to querystring query,
    # and those with fields specified to typical match queries.

    def __init__(
        self,
        user_query=None,  # like a prepared statement in SQL
        scopes_regexs=(),  # inference used when encountering empty scopes
        scopes_default=("_id",),  # fallback used when scope inference fails
        allow_random_query=True,  # used for data exploration, can be expensive
        allow_nested_query=False,  # nested aggregation can be expensive
        metadata=None,  # access to data like total number of documents
    ):
        # for autoscope feature, to infer scope from q when enabled
        self.parser = QStringParser(scopes_default, scopes_regexs)

        # all settings below affect only query string queries
        self.user_query = user_query or ESUserQuery("userquery")
        self.allow_random_query = allow_random_query
        self.allow_nested_query = allow_nested_query  # for aggregations

        # currently metadata is only used for __any__ query
        self.metadata = metadata

    def build(self, q=None, **options):
        """
        Build a query according to q and options.
        This is the public method called by API handlers.

        Regarding scopes:
            scopes: [str] nonempty, match query.
            scopes: NoneType, or [], no scope, so query string query.

        Additionally support these options:
            explain: include es scoring information
            userquery: customized function to interpret q

        * additional keywords are passed through as es keywords
            for example: 'explain', 'version' ...

        * multi-search is supported when q is a list. all queries
            are built individually and then sent in one request.

        """
        options = dotdict(options)

        if options.scroll_id:
            # bypass all query building stages
            return ESScrollID(options.scroll_id)

        if options.fetch_all:
            # clean up conflicting parameters
            options.pop("sort", None)
            options.pop("size", None)

        try:
            # process single q vs list of q(s).
            # dispatch 'val' vs 'key:val' to corresponding functions.

            if isinstance(q, list):
                search = MultiSearch()
                for _q in q:
                    _search = self._build_one(_q, options)
                    search = search.add(_search)
            else:  # str, int ...
                search = self._build_one(q, options)

        except IllegalOperation as exc:
            raise ValueError(str(exc))  # ex. sorting by -_score

        if options.get("rawquery"):
            raise RawQueryInterrupt(search.to_dict())

        return search

    def _build_one(self, q, options):
        # a single query, possibly included in a multi-search
        # later but it itself is a single query unit.

        if options.scopes:
            search = self._build_match_query(q, options.scopes, options)
        elif not isinstance(q, (list, tuple)) and options.autoscope:
            q, scopes = self.parser.parse(str(q))
            search = self._build_match_query(q, scopes, options)
        else:  # no scope provided and cannot derive from q
            search = self._build_string_query(q, options)

        # pass through es query options. (from, size ...)
        search = self.apply_extras(search, options)
        return search

    def _build_string_query(self, q, options):
        """q + options -> query object

        options:
            userquery
        """
        search = Search()
        userquery = options.userquery or ""

        if q == "":  # same empty q behavior as that of ES.
            search = search.query("match_none")

        elif q == "__all__" or q is None:
            search = search.query()
            if options.aggs and not options.size:
                options.size = 0

        elif q == "__any__":
            if self.allow_random_query:
                search = search.query("function_score", random_score={})
            else:  # pseudo random by overriding 'from' value
                search = search.query()
                try:  # limit 'from' parameter to a valid result window
                    metadata = self.metadata[options.biothing_type]
                    total = metadata["stats"]["total"]
                    fmax = total - options.get("size", 0)
                    from_ = randrange(fmax if fmax < 10000 else 10000)
                    options["from"] = from_ if from_ >= 0 else 0
                except Exception:
                    raise ValueError("random query not available.")

        elif self.user_query.has_query(userquery):
            userquery_ = self.user_query.get_query(userquery, q=q)
            search = search.query(userquery_)

        else:  # customization here
            search = self.default_string_query(str(q), options)

        if self.user_query.has_filter(userquery):
            userfilter = self.user_query.get_filter(userquery)
            search = search.filter(userfilter)

        return search

    def _build_match_query(self, q, scopes, options):
        """q + scopes + options -> query object

        case 1:
            # single match query
            q = "1017"
            scopes = ["_id"] or "_id"

        case 2:
            # compound match query
            q = ["1017", "CDK2"]
            scopes = [["_id", "entrezgene"], "symbol"]
        """

        if not isinstance(q, (list, tuple)):
            q, scopes = [q], [scopes]

        # considering the complexity of data types,
        # for example, q can take the type of int, bool, and float,
        # maybe it's better to let elasticsearch or its python package
        # handle the type checking. the checks below represent a
        # typical case but is inconclusive.

        # if not all((
        #         isinstance(q, (list, tuple)),
        #         all(isinstance(_q, str) for _q in q))):
        #     raise TypeError("Expect q: Union[list[str], str].", q)

        # if not all((
        #         isinstance(scopes, (list, tuple)),
        #         all(isinstance(_s, (list, tuple, str)) for _s in scopes))):
        #     raise TypeError("Expect scopes: list[Union[list[str], str]].", scopes)

        if not len(q) == len(scopes):
            raise ValueError("Expect len(q) == len(scopes).")

        # additional uncommon type errors
        # will be raised in elasticsearch

        search = Search()
        for _q, _scopes in zip(q, scopes):
            if not (_q and _scopes):
                raise ValueError("No search terms or scopes.")
            _search = self.default_match_query(_q, _scopes, options)
            search = search.query(_search.query)
        return search

    def default_string_query(self, q, options):
        """
        Override this to customize default string query.
        By default it implements a query string query.
        """
        assert isinstance(q, str) and q
        assert not options.scopes
        return Search().query("query_string", query=q, default_operator="AND", lenient=True)

    def default_match_query(self, q, scopes, options):
        """
        Override this to customize default match query.
        By default it implements a multi_match query.
        """
        assert isinstance(q, (str, int, float, bool))
        assert isinstance(scopes, (list, tuple, str)) and scopes
        _params = dict(query=q, fields=scopes, operator="AND", lenient=True)
        if options.analyzer:
            _params["analyzer"] = options.analyzer
        return Search().query("multi_match", **_params)

    def apply_extras(self, search, options):
        """
        Process non-query options and customize their behaviors.
        Customized aggregation syntax string is translated here.
        """

        # add aggregations
        facet_size = options.facet_size or 10
        for agg in options.aggs or []:
            term, bucket = agg, search.aggs
            while term:
                if self.allow_nested_query and "(" in term and term.endswith(")"):
                    _term, term = term[:-1].split("(", 1)
                else:
                    _term, term = term, ""
                bucket = bucket.bucket(_term, "terms", field=_term, size=facet_size)

        # add es params
        if isinstance(options.sort, list):
            # accept '-' prefixed field names
            search = search.sort(*options.sort)
        if isinstance(options._source, list):
            if "all" not in options._source:
                fields_with_minus = [field.lstrip("-") for field in options._source if field.startswith("-")]
                fields_without_minus = [field for field in options._source if not field.startswith("-")]
                search = search.source(includes=fields_without_minus, excludes=fields_with_minus)
        for key in ("from", "size", "explain", "version"):
            if key in options:
                search = search.extra(**{key: options[key]})

        # the valid values for from and size depend on the
        # index.max_result_window elasticsearch setting.

        # more about this constraint on:
        # https://www.elastic.co/guide/en/elasticsearch/
        # reference/current/index-modules.html

        # Feature: filter
        # apply extra filter (as query_string query) to filter results
        # Ref: https://www.elastic.co/guide/en/elasticsearch/reference/8.10/query-dsl-bool-query.html
        if options.filter:
            search = search.filter("query_string", query=options.filter)

        # Feature: post_filter
        # -- implementation using query string matching
        # Ref: https://www.elastic.co/guide/en/elasticsearch/reference/8.10/filter-search-results.html#post-filter
        if options.post_filter:
            search = search.post_filter("query_string", query=options["post_filter"])

        return search


class MongoQueryBuilder:
    def __init__(self, default_scopes=("_id",)):
        self.parser = QStringParser(default_scopes)

    def build(self, q, **options):
        fields = options.get("scopes", ())
        if not fields and q:
            q, fields = self.parser.parse(q)

        assert isinstance(fields, (list, tuple))
        assert q is None and not fields or q and isinstance(q, str)
        assert all((isinstance(field, str) for field in fields))

        filter_ = {field: 1 for field in options.get("_source", ())} or None  # project fields to return

        query = {"$or": [{field: q} for field in fields]} if fields else {}

        if options.get("rawquery"):
            raise RawQueryInterrupt((query, filter_))

        return (query, filter_)


class SQLQueryBuilder:
    # PROOF OF CONCEPT
    # INPUT NOT SANITIZED
    # INTERNAL USE ONLY

    def __init__(self, tables, default_scopes=("id",), default_limit=10):
        assert default_scopes
        assert isinstance(default_limit, int)
        assert tables and isinstance(tables, dict)

        self.tables = tables
        self.default_limit = default_limit
        self.parser = QStringParser(default_scopes)

        if None not in self.tables:  # set default table
            self.tables[None] = next(iter(self.tables.values()))

    def build(self, q, **options):
        statements = [
            "SELECT {}".format(", ".join(options.get("_source", ())) or "*"),
            "FROM {}".format(self.tables[options.get("biothing_type")]),
        ]

        scopes = options.get("scopes")
        if not scopes:
            q, scopes = self.parser.parse(q)

        if scopes and q:
            assert isinstance(q, str)
            selections = ['{} = "{}"'.format(field, q) for field in scopes]
            statements.append("WHERE")
            statements.append(" OR ".join(selections))

        # limit result window
        statements.append("LIMIT {}".format(options.get("size", self.default_limit)))

        if "from_" in options:
            statements.append("OFFSET {}".format(options["from_"]))

        if options.get("rawquery"):
            raise RawQueryInterrupt(statements)

        return " ".join(statements)


class ESUserQuery:
    def __init__(self, path):
        self._queries = {}
        self._filters = {}
        try:
            for dirpath, dirnames, filenames in os.walk(path):
                if dirnames:
                    self.logger.info("User query folders: %s.", dirnames)
                    continue
                for filename in filenames:
                    with open(os.path.join(dirpath, filename)) as text_file:
                        if "query" in filename:
                            ## alternative implementation  # noqa: E266
                            # self._queries[os.path.basename(dirpath)] = text_file.read()
                            ##
                            self._queries[os.path.basename(dirpath)] = orjson.loads(text_file.read())
                        elif "filter" in filename:
                            self._filters[os.path.basename(dirpath)] = orjson.loads(text_file.read())
        except Exception:
            self.logger.exception("Error loading user queries.")

    def has_query(self, named_query):
        return named_query in self._queries

    def has_filter(self, named_query):
        return named_query in self._filters

    def get_query(self, named_query, **kwargs):
        def in_place_sub(dic, kwargs):
            for key in dic:
                if isinstance(dic[key], dict):
                    in_place_sub(dic[key], kwargs)
                elif isinstance(dic[key], list):
                    for item in dic[key]:
                        in_place_sub(item, kwargs)
                elif isinstance(dic[key], str):
                    dic[key] = dic[key].format(**kwargs).format(**kwargs)  # {{q}}

        dic = deepcopy(self._queries.get(named_query))
        in_place_sub(dic, kwargs)
        key, val = next(iter(dic.items()))
        return Q(key, **val)

        ## alternative implementation  # noqa: E266
        # string = self._queries.get(named_query)
        # string1 = re.sub(r"\}", "}}", string)
        # string2 = re.sub(r"\{", "{{", string1)
        # string3 = re.sub(r'\{\{\{\{(?P<var>.*?)\}\}\}\}', r'{\g<var>}', string2)
        # return string3
        ##

    def get_filter(self, named_query):
        dic = self._filters.get(named_query)
        key, val = next(iter(dic.items()))
        return Q(key, **val)

    @property
    def logger(self):
        return logging.getLogger(__name__)



import sys
from contextlib import ContextDecorator
from io import StringIO


class RedirectStdStreams(ContextDecorator):
    def __init__(self, stdout=None, stderr=None):
        self._stdout = stdout or StringIO()
        self._stderr = stderr or StringIO()

    def __enter__(self):
        self.old_stdout, self.old_stderr = sys.stdout, sys.stderr
        self.old_stdout.flush()
        self.old_stderr.flush()
        sys.stdout, sys.stderr = self._stdout, self._stderr
        return self

    def __exit__(self, *exc):
        self._stdout.flush()
        self._stderr.flush()
        sys.stdout = self.old_stdout
        sys.stderr = self.old_stderr

    def get_std_contents(self):
        self._stdout.seek(0)
        self._stderr.seek(0)
        stdout_content = "".join([line for line in self._stdout.readlines()])
        stderr_content = "".join([line for line in self._stderr.readlines()])
        return {
            "stdout": stdout_content,
            "stderr": stderr_content,
        }



import json
import logging
import os
import sys

from biothings.utils.version import get_software_info

logger = logging.getLogger(__name__)


class FieldNote:
    def __init__(self, path):
        try:  # populate field notes if exist
            inf = open(path, "r")
            self._fields_notes = json.load(inf)
            inf.close()
        except Exception:
            self._fields_notes = {}

    def get_field_notes(self):
        """
        Return the cached field notes associated with this instance.
        """
        return self._fields_notes


class DevInfo:
    def __init__(self):
        # the path can be a sub-folder of a git repository
        # as long as it responds to "git" command line call

        if sys.argv[0]:  # running as a script
            self._repo = os.path.dirname(sys.argv[0])
        else:  # REPL
            self._repo = os.getcwd()

    def get(self):
        # {
        #     "python-package-info": [
        #         "aiocron                 1.6",
        #         "aiohttp                 3.6.2",
        #         "async-timeout           3.0.1",
        #         ...
        #     ],
        #     "codebase": {
        #         "repository-url": "https://github.com/biothings/biothings.api.git",
        #         "commit-hash": "8715a8ebcf49bc436bca24b1ce1a35f1777a7c4c"
        #     },
        #     "biothings": {
        #         "repository-url": "https://github.com/biothings/biothings.api.git",
        #         "commit-hash": "f3ca99a1f5261a70eaef196bbab1163302c16859",
        #         "master-commits": "2478",
        #         "version": "0.10.0"
        #     },
        #     "python-info": {
        #         "version": "3.9.5 (tags/v3.9.5:0a7dcbd, May  3 2021, 17:27:52)",
        #         "version_info": {
        #             "major": 3,
        #             "minor": 9,
        #             "micro": 5
        #         }
        #     }
        # }

        try:
            return get_software_info(self._repo)
        except Exception:
            return {}



"""
Biothings Web Handlers

biothings.web.handlers.BaseHandler

    Supports:
    - access to biothings namespace
    - monitor exceptions with Sentry

biothings.web.handlers.BaseAPIHandler

    Additionally supports:
    - JSON and YAML payload in the request body
    - request arguments standardization
    - multi-type output (json, yaml, html, msgpack)
    - standardized error response (exception -> error template)
    - analytics and usage tracking (Google Analytics and AWS)
    - default common http headers (CORS and Cache Control)

"""
import logging

import orjson
import yaml
from tornado.web import HTTPError, RequestHandler

from biothings.utils import serializer
from biothings.web.analytics.events import Event
from biothings.web.analytics.notifiers import AnalyticsMixin
from biothings.web.options import OptionError, ReqArgs

logger = logging.getLogger(__name__)


class BaseHandler(RequestHandler):
    @property
    def biothings(self):
        return self.application.biothings


class BaseAPIHandler(BaseHandler, AnalyticsMixin):
    name = "__base__"
    kwargs = {
        "*": {
            "format": {
                "type": str,
                "default": "json",
                "enum": ("json", "yaml", "html", "msgpack"),
            }
        }
    }
    format = "json"
    cache = None
    cache_control_template = "max-age={cache}, public"

    def initialize(self, cache=None):
        cache_value = self.biothings.config.DEFAULT_CACHE_MAX_AGE
        if self.cache is not None:
            cache_value = self.cache
        if cache is not None:
            cache_value = cache
        # self._header has already set when call set_default_headers func before
        # so we need to overwrite it to make custom cache age works
        self.set_cache_header(cache_value)

        self.args = {}  # processed args will be available here
        self.args_query = {}  # query parameters in the URL
        self.args_form = {}  # form-data and x-www-form-urlencoded
        self.args_json = {}  # applicatoin/json type body
        self.args_yaml = {}  # applicatoin/yaml type body
        self.event = Event()

        # do not assume the data types of some the variables
        # defined above. self.args can be a dotdict after
        # processing. json/yaml can be any serializable objs.
        # self.event may be replaced with its sub-classes.

    def prepare(self):
        content_type = self.request.headers.get("Content-Type", "")
        if content_type.startswith("application/json"):
            self.args_json = self._parse_json()
        elif content_type.startswith("application/yaml"):
            self.args_yaml = self._parse_yaml()

        self.args_query = {key: self.get_query_argument(key) for key in self.request.query_arguments}
        self.args_form = {key: self.get_body_argument(key) for key in self.request.body_arguments}

        reqargs = ReqArgs(
            ReqArgs.Path(args=self.path_args, kwargs=self.path_kwargs),
            query=self.args_query,
            form=self.args_form,
            json_=self.args_json,
        )
        # standardized request arguments
        self.args = self._parse_args(reqargs)
        self.format = self.args.format

    def _parse_json(self):
        if not self.request.body:
            raise HTTPError(
                400,
                reason=(
                    "Empty body is not a valid JSON. "
                    "Remove the content-type header, or "
                    "provide an empty object in the body."
                ),
            )
        try:
            return orjson.loads(self.request.body)
        except orjson.JSONDecodeError:
            raise HTTPError(400, reason="Invalid JSON body.")

    def _parse_yaml(self):
        try:
            return yaml.load(self.request.body, Loader=yaml.SafeLoader)
        except (yaml.scanner.ScannerError, yaml.parser.ParserError):
            raise HTTPError(400, reason="Invalid YAML body.")

    def _parse_args(self, reqargs):
        if not self.name:  # feature disabled
            return {}  # default value

        optionsets = self.biothings.optionsets
        optionset = optionsets.get(self.name)

        try:  # uses biothings.web.options to standardize args
            args = optionset.parse(self.request.method, reqargs)

        except OptionError as err:
            args = err  # for logging in "finally" clause
            raise HTTPError(400, None, err.info)

        else:  # set on self.args
            return args

        finally:  # one log message regardless of success
            logger.debug("%s %s\n%s\n%s", self.request.method, self.request.uri, reqargs, args)

    def write(self, chunk):
        try:
            if self.format == "json":
                chunk = serializer.to_json(chunk)
                self.set_header("Content-Type", "application/json; charset=UTF-8")

            elif self.format == "yaml":
                chunk = serializer.to_yaml(chunk)
                self.set_header("Content-Type", "text/x-yaml; charset=UTF-8")

            elif self.format == "msgpack":
                chunk = serializer.to_msgpack(chunk)
                self.set_header("Content-Type", "application/x-msgpack")

            elif self.format == "html":
                chunk = self.render_string("api.html", data=serializer.to_json(chunk))
                self.set_header("Content-Type", "text/html; charset=utf-8")

        except Exception as exc:
            # this is a low-level method, used in many places,
            # error handling should happen in the upper layers,
            logger.warning(exc)

        super().write(chunk)

    def get_template_path(self):
        # APIs should not normally need to use templating
        # set the path to where we can find the api.html
        import biothings.web.templates

        return next(iter(biothings.web.templates.__path__))

    def on_finish(self):
        """
        This is a tornado lifecycle hook.
        Override to provide tracking features.
        """
        logger.debug(self.event)
        super().on_finish()

    def write_error(self, status_code, **kwargs):
        """
        from tornado.web import Finish, HTTPError

        raise HTTPError(404)
        raise HTTPError(404, reason="document not found")
        raise HTTPError(404, None, {"id": "-1"}, reason="document not found") ->
        {
          "code": 404,
          "success": False,
          "error": "document not found"
          "id": "-1"
        }
        """

        reason = kwargs.pop("reason", self._reason)
        # "reason" is a reserved tornado keyword
        # see RequestHandler.send_error
        assert isinstance(reason, str)
        assert "\n" not in reason

        message = {"code": status_code, "success": False, "error": reason}
        try:  # merge exception info
            exception = kwargs["exc_info"][1]
            if isinstance(exception.args[0], dict):
                message.update(exception.args[0])
        except Exception:
            pass

        self.finish(message)

    def options(self, *args, **kwargs):
        self.set_status(204)
        self.finish()

    def set_default_headers(self):
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header("Access-Control-Allow-Methods", "*")
        self.set_header("Access-Control-Allow-Headers", "*")
        self.set_header("Access-Control-Allow-Credentials", "false")
        self.set_header("Access-Control-Max-Age", "60")

    def set_cache_header(self, cache_value):
        if isinstance(cache_value, int):
            # https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Cache-Control
            # to disable caching for a handler, set cls.cache to 0 or
            # run self.clear_header('Cache-Control') in an HTTP method
            # or set cache value on the config file:
            # r"/api/query/?", "biothings.web.handlers.QueryHandler", {"biothing_type": "schema", "cache": 0}),
            self.set_header("Cache-Control", self.cache_control_template.format(cache=cache_value))



"""
    The "templates" folder stores HTML templates for native biothings
    web handlers and is structured as a dummy module to facilitate location
    resolution. Typically, ``biothings.web.templates.__path__[0]`` returns
    a string representing the location of this folder on the file system.
    This folder is intended to be used internally by the SDK developer.
"""



import abc
import json
from typing import Iterable, Optional, Tuple, Type

from biothings.web.handlers import BaseAPIHandler

__all__ = [
    "BioThingsAuthenticationProviderInterface",
    "BioThingsAuthnMixin",
]


class BioThingsAuthenticationProviderInterface(abc.ABC):
    """
    Authentication Provider Interface for BioThings API Endpoints

    BioThingsAuthnMixin depends on this interface. Any authentication for
    the API endpoints is recommended to be implemented using this
    interface.

    A provider will be initialized with the API Handler calling it, so it will
    have access to the request, and other related BioThings facilities.

    Args:
        handler: the BaseAPIHandler that invoked this provider

    Attributes:
        WWW_AUTHENTICATE_HEADER: string used for the 'WWW-Authenticate' Header.
            If the handler returns a 401, this may be used to populate the
            'WWW-Authenticate' Header. It should be None or one from the list
            https://www.iana.org/assignments/http-authschemes/http-authschemes.xhtml
    """

    WWW_AUTHENTICATE_HEADER: Optional[str] = None

    def __init__(self, handler: BaseAPIHandler, **kwargs):
        # we use the BioThings BaseAPIHandler so that:
        #  - the interface has access to self.handler.biothings
        #  - it is something only intended for programmatic access
        super().__init__()
        self.handler = handler

    @abc.abstractmethod
    def get_current_user(self) -> Optional[dict]:
        """
        Get the current user.

        If authentication succeeds, return the current user.
        If authentication fails, return None
        """
        raise NotImplementedError


class BioThingsAuthnMixin(BaseAPIHandler):
    """
    MixIn Class to for Pluggable User Authentication

    Configuring `AUTHN_PROVIDERS` setting to a list of tuples, each tuple
    containing an authentication provider and its additional initialization
    arguments.

    The authentication providers will be processed in the order they are
    configured.

    Mixing in this class enables getting the current authenticated user through
    the `current_user` attribute of RequestHandler instances.

    The `get_www_authenticate_header` will offer a value to populate the
    WWW-Authenticate header, in the case that a 401 status code needs to be returned.

    Example:
        >>> from tornado.web import HTTPError
        >>> class ExampleGetUserInfoHandler(BioThingsAuthnMixin, BaseAPIHandler):
        >>>     def get(self):
        >>>         # if the current user is set, send it to output
        >>>         if self.current_user:
        >>>             self.write(self.current_user)
        >>>         else:
        >>>             # not authenticated, return 401 or 403
        >>>             header = self.get_www_authenticate_header()
        >>>             if header:
        >>>                 self.clear()
        >>>                 self.set_header('WWW-Authenticate', header)
        >>>                 self.set_status(401, "Unauthorized")
        >>>                 # raising HTTPError will cause headers to be emptied
        >>>                 self.finish()
        >>>             else:
        >>>                 raise HTTPError(403)

    Attributes:
        AUTHN_PROVIDERS: Overrides the global options for the specific handler.
            Do not set this attribute if not intending to override the default.

    Notes:
        Beware of MRO issues, see example for proper order of multi-inheritance.
    """

    def get_current_user(self):
        """
        Get the user from list of preconfigured authentication providers.
        """
        # Support pluggable authentication.
        # Compare to PAM in Linux. Sample logic below
        authenticators: Iterable[Tuple[Type[BioThingsAuthenticationProviderInterface], dict]] = getattr(
            self, "AUTHN_PROVIDERS", self.biothings.config.AUTHN_PROVIDERS
        )

        # loop through the list in order and initialize the provider using
        # self & configured options (like how handlers are configured for routing)
        # use the provider to check if the user can log in
        for authenticator_cls, init_kwargs in authenticators:
            authenticator = authenticator_cls(self, **init_kwargs)
            user = authenticator.get_current_user()
            if user:
                return user
        return None

    def get_www_authenticate_header(self) -> Optional[str]:
        """
        Get the most preferred header to populate WWW-Authenticate

        If this method returns None, it is better to set a 403 error status code.
        Otherwise, return a 401 error and use the value returned in the
        WWW-Authenticate header.

        According to RFC 7235 https://datatracker.ietf.org/doc/html/rfc7235#section-3.1
        "the server generating a 401 response MUST send a WWW-Authenticate header field".
        """
        authenticators: Iterable[Tuple[Type[BioThingsAuthenticationProviderInterface], dict], ...] = getattr(
            self, "AUTHN_PROVIDERS", self.biothings.config.AUTHN_PROVIDERS
        )
        for authenticator_cls, _ in authenticators:
            header = authenticator_cls.WWW_AUTHENTICATE_HEADER
            if header is not None:
                return header
        return None


class DefaultCookieAuthnProvider(BioThingsAuthenticationProviderInterface):
    WWW_AUTHENTICATE_HEADER = "None"

    def __init__(self, handler, cookie_name="user"):
        super(DefaultCookieAuthnProvider, self).__init__(handler)
        self.cookie_name = cookie_name

    def get_current_user(self):
        user = self.handler.get_secure_cookie(self.cookie_name)
        if not user:
            return None
        return json.loads(user.decode())
